import math
import typing
from abjad import Fraction
from abjad import exceptions
from abjad import mathtools
from abjad.lilypondnames.LilyPondTweakManager import LilyPondTweakManager
from abjad.mathtools.NonreducedFraction import NonreducedFraction
from abjad.mathtools.NonreducedRatio import NonreducedRatio
from abjad.mathtools.Ratio import Ratio
from abjad.system.FormatSpecification import FormatSpecification
from abjad.top.inspect import inspect
from abjad.top.iterate import iterate
from abjad.top.override import override
from abjad.top.tweak import tweak
from abjad.utilities.Duration import Duration
from abjad.utilities.Multiplier import Multiplier
from .Container import Container
from .Leaf import Leaf
from .LeafMaker import LeafMaker
from .Measure import Measure
from .Note import Note
from .NoteMaker import NoteMaker
from .Rest import Rest


class Tuplet(Container):
    r"""
    Tuplet.

    ..  container:: example

        A tuplet:

        >>> tuplet = abjad.Tuplet((2, 3), "c'8 d'8 e'8")
        >>> abjad.show(tuplet) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(tuplet)
            \times 2/3 {
                c'8
                d'8
                e'8
            }

    ..  container:: example

        A nested tuplet:

        >>> second_tuplet = abjad.Tuplet((4, 7), "g'4. ( a'16 )")
        >>> tuplet.insert(1, second_tuplet)
        >>> abjad.show(tuplet) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(tuplet)
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                c'8
                \times 4/7 {
                    g'4.
                    (
                    a'16
                    )
                }
                d'8
                e'8
            }


    ..  container:: example

        A doubly nested tuplet:

            >>> third_tuplet = abjad.Tuplet((4, 5), [])
            >>> third_tuplet.extend("e''32 [ ef''32 d''32 cs''32 cqs''32 ]")
            >>> second_tuplet.insert(1, third_tuplet)
            >>> abjad.show(tuplet) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(tuplet)
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                c'8
                \tweak edge-height #'(0.7 . 0)
                \times 4/7 {
                    g'4.
                    (
                    \times 4/5 {
                        e''32
                        [
                        ef''32
                        d''32
                        cs''32
                        cqs''32
                        ]
                    }
                    a'16
                    )
                }
                d'8
                e'8
            }

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Containers'

    __slots__ = (
        '_denominator',
        '_force_fraction',
        '_hide',
        '_multiplier',
        '_tweaks',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        multiplier=(2, 3),
        components=None,
        *,
        denominator: int = None,
        force_fraction: bool = None,
        hide: bool = None,
        tweaks: typing.Union[
            typing.List[typing.Tuple], LilyPondTweakManager] = None,
        ) -> None:
        Container.__init__(self, components)
        multiplier = Multiplier(multiplier)
        self.multiplier = multiplier
        self.denominator = denominator
        self.force_fraction = force_fraction
        self.hide = hide
        self._tweaks = None
        LilyPondTweakManager.set_tweaks(self, tweaks)

    ### SPECIAL METHODS ###

    def __getnewargs__(self) -> tuple:
        """
        Gets new arguments of tuplet.
        """
        return (self.multiplier,)

    ### PRIVATE METHODS ###

    def _as_graphviz_node(self):
        node = Container._as_graphviz_node(self)
        node[0].extend([
            uqbar.graphs.TableRow([
                uqbar.graphs.TableCell(
                    label=type(self).__name__,
                    attributes={'border': 0},
                    ),
                ]),
            uqbar.graphs.HRule(),
            uqbar.graphs.TableRow([
                uqbar.graphs.TableCell(
                    label=f'* {self.multiplier!s}',
                    attributes={'border': 0},
                    ),
                ]),
            ])
        return node

    def _format_after_slot(self, bundle):
        result = []
        result.append(('grob reverts', bundle.grob_reverts))
        result.append(('commands', bundle.after.commands))
        result.append(('comments', bundle.after.comments))
        return tuple(result)

    def _format_before_slot(self, bundle):
        result = []
        result.append(('comments', bundle.before.comments))
        result.append(('commands', bundle.before.commands))
        result.append(('grob overrides', bundle.grob_overrides))
        result.append(('context settings', bundle.context_settings))
        return tuple(result)

    def _format_close_brackets_slot(self, bundle):
        result = []
        if self.multiplier:
            result.append([('self_brackets', 'close'), '}'])
        return tuple(result)

    def _format_closing_slot(self, bundle):
        result = []
        result.append(('commands', bundle.closing.commands))
        result.append(('comments', bundle.closing.comments))
        return self._format_slot_contributions_with_indent(result)

    def _format_lilypond_fraction_command_string(self):
        if self.hide:
            return ''
        if 'text' in vars(override(self).tuplet_number):
            return ''
        if (self.augmentation() or
            not self._get_power_of_two_denominator() or
            self.multiplier.denominator == 1 or
            self.force_fraction
            ):
            return r"\tweak text #tuplet-number::calc-fraction-text"
        return ''

    def _format_open_brackets_slot(self, bundle):
        result = []
        if self.multiplier:
            if self.hide:
                contributor = (self, 'hide')
                scale_durations_command_string = \
                    self._get_scale_durations_command_string()
                contributions = [scale_durations_command_string]
                result.append([contributor, contributions])
            else:
                contributor = ('self_brackets', 'open')
                contributions = []
                fraction_command_string = \
                    self._format_lilypond_fraction_command_string()
                if fraction_command_string:
                    contributions.append(fraction_command_string)
                edge_height_tweak_string = \
                    self._get_edge_height_tweak_string()
                if edge_height_tweak_string:
                    contributions.append(edge_height_tweak_string)

                strings = tweak(self)._list_format_contributions(
                    directed=False,
                    )
                contributions.extend(strings)

                times_command_string = self._get_times_command_string()
                contributions.append(times_command_string)
                result.append([contributor, contributions])
        return tuple(result)

    def _format_opening_slot(self, bundle):
        result = []
        result.append(('comments', bundle.opening.comments))
        result.append(('commands', bundle.opening.commands))
        return self._format_slot_contributions_with_indent(result)

    def _get_compact_representation(self):
        if not self:
            return f'{{ {self.multiplier!s} }}'
        return f'{{ {self.multiplier!s} {self._get_contents_summary()} }}'

    def _get_edge_height_tweak_string(self):
        measure = inspect(self).parentage().get_first(Measure)
        if measure and measure.implicit_scaling:
            return
        duration = self._get_preprolated_duration()
        denominator = duration.denominator
        if not mathtools.is_nonnegative_integer_power_of_two(denominator):
            return r"\tweak edge-height #'(0.7 . 0)"

    def _get_format_specification(self):
        return FormatSpecification(
            client=self,
            repr_args_values=[self.multiplier, self._get_contents_summary()],
            storage_format_args_values=[self.multiplier, self[:]],
            storage_format_kwargs_names=[],
            )

    def _get_lilypond_format(self):
        self._update_now(indicators=True)
        return self._format_component()

    def _get_multiplier_fraction_string(self):
        if self.denominator is not None:
            inverse_multiplier = Multiplier(
                self.multiplier.denominator, self.multiplier.numerator)
            nonreduced_fraction = NonreducedFraction(inverse_multiplier)
            nonreduced_fraction = nonreduced_fraction.with_denominator(
                self.denominator)
            denominator, numerator = nonreduced_fraction.pair
        else:
            numerator = self.multiplier.numerator
            denominator = self.multiplier.denominator
        return f'{numerator}/{denominator}'

    def _get_power_of_two_denominator(self):
        if self.multiplier:
            numerator = self.multiplier.numerator
            return mathtools.is_nonnegative_integer_power_of_two(numerator)
        else:
            return True

    def _get_preprolated_duration(self):
        return self.multiplied_duration

    def _get_ratio_string(self):
        multiplier = self.multiplier
        if multiplier is not None:
            numerator = multiplier.numerator
            denominator = multiplier.denominator
            ratio_string = f'{denominator}:{numerator}'
            return ratio_string
        else:
            return None

    def _get_scale_durations_command_string(self):
        multiplier = self.multiplier
        numerator = multiplier.numerator
        denominator = multiplier.denominator
        string = rf"\scaleDurations #'({numerator} . {denominator}) {{"
        return string

    def _get_summary(self):
        if 0 < len(self):
            return ', '.join([str(x) for x in self.components])
        else:
            return ''

    def _get_times_command_string(self):
        string = rf'\times {self._get_multiplier_fraction_string()} {{'
        return string

    def _scale(self, multiplier):
        multiplier = Multiplier(multiplier)
        for component in self[:]:
            if isinstance(component, Leaf):
                new_duration = multiplier * component.written_duration
                component._set_duration(new_duration)
        self.normalize_multiplier()

    ### PUBLIC PROPERTIES ###

    @property
    def denominator(self) -> typing.Optional[int]:
        r"""
        Gets and sets preferred denominator of tuplet.

        ..  container:: example

            Gets preferred denominator of tuplet:

            >>> tuplet = abjad.Tuplet((2, 3), "c'8 d'8 e'8")
            >>> tuplet.denominator is None
            True
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 2/3 {
                    c'8
                    d'8
                    e'8
                }

        ..  container:: example

            Sets preferred denominator of tuplet:

            >>> tuplet = abjad.Tuplet((2, 3), "c'8 d'8 e'8")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 2/3 {
                    c'8
                    d'8
                    e'8
                }

            >>> tuplet.denominator = 4
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 4/6 {
                    c'8
                    d'8
                    e'8
                }

        """
        return self._denominator

    @denominator.setter
    def denominator(self, argument):
        if isinstance(argument, int):
            if not 0 < argument:
                raise ValueError(argument)
        elif not isinstance(argument, type(None)):
            raise TypeError(argument)
        self._denominator = argument

    @property
    def force_fraction(self) -> typing.Optional[bool]:
        r"""
        Gets and sets force fraction flag.

        ..  container:: example

            The ``default.ily`` stylesheet included in all Abjad API examples
            includes the following:
            
            ``\override TupletNumber.text = #tuplet-number::calc-fraction-text``

            This means that even simple tuplets format as explicit fractions:

            >>> staff = abjad.Staff()
            >>> staff.append(abjad.Tuplet((2, 3), "c'4 d' e'"))
            >>> staff.append(abjad.Tuplet((2, 3), "c'4 d' e'"))
            >>> staff.append(abjad.Tuplet((2, 3), "c'4 d' e'"))
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \times 2/3 {
                        c'4
                        d'4
                        e'4
                    }
                    \times 2/3 {
                        c'4
                        d'4
                        e'4
                    }
                    \times 2/3 {
                        c'4
                        d'4
                        e'4
                    }
                }

            To illustrate the effect of Abjad's force fraction property, we can
            temporarily restore LilyPond's default tuplet number formatting
            like this:

            >>> staff = abjad.Staff()
            >>> staff.append(abjad.Tuplet((2, 3), "c'4 d' e'"))
            >>> staff.append(abjad.Tuplet((2, 3), "c'4 d' e'"))
            >>> staff.append(abjad.Tuplet((2, 3), "c'4 d' e'"))
            >>> string = 'tuplet-number::calc-denominator-text'
            >>> abjad.override(staff).tuplet_number.text = string
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                \with
                {
                    \override TupletNumber.text = #tuplet-number::calc-denominator-text
                }
                {
                    \times 2/3 {
                        c'4
                        d'4
                        e'4
                    }
                    \times 2/3 {
                        c'4
                        d'4
                        e'4
                    }
                    \times 2/3 {
                        c'4
                        d'4
                        e'4
                    }
                }

            Which makes it possible to see the effect of setting force fraction
            to true on a single tuplet:

            >>> tuplet = staff[1]
            >>> tuplet.force_fraction = True
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                \with
                {
                    \override TupletNumber.text = #tuplet-number::calc-denominator-text
                }
                {
                    \times 2/3 {
                        c'4
                        d'4
                        e'4
                    }
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 2/3 {
                        c'4
                        d'4
                        e'4
                    }
                    \times 2/3 {
                        c'4
                        d'4
                        e'4
                    }
                }

        ..  container:: example

            Ignored when tuplet number text is overridden explicitly:

            >>> tuplet = abjad.Tuplet((2, 3), "c'8 d'8 e'8")
            >>> duration = abjad.inspect(tuplet).duration()
            >>> markup = duration.to_score_markup()
            >>> abjad.override(tuplet).tuplet_number.text = markup
            >>> staff = abjad.Staff([tuplet])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \override TupletNumber.text = \markup {
                        \score
                            {
                                \new Score
                                \with
                                {
                                    \override SpacingSpanner.spacing-increment = #0.5
                                    proportionalNotationDuration = ##f
                                }
                                <<
                                    \new RhythmicStaff
                                    \with
                                    {
                                        \remove Time_signature_engraver
                                        \remove Staff_symbol_engraver
                                        \override Stem.direction = #up
                                        \override Stem.length = #5
                                        \override TupletBracket.bracket-visibility = ##t
                                        \override TupletBracket.direction = #up
                                        \override TupletBracket.minimum-length = #4
                                        \override TupletBracket.padding = #1.25
                                        \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                        \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                        \override TupletNumber.font-size = #0
                                        \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                        tupletFullLength = ##t
                                    }
                                    {
                                        c'4
                                    }
                                >>
                                \layout {
                                    indent = #0
                                    ragged-right = ##t
                                }
                            }
                        }
                    \times 2/3 {
                        c'8
                        d'8
                        e'8
                    }
                    \revert TupletNumber.text
                }

        """
        return self._force_fraction

    @force_fraction.setter
    def force_fraction(self, argument):
        if isinstance(argument, (bool, type(None))):
            self._force_fraction = argument
        else:
            message = f'force fraction must be boolean (not {argument!r}).'
            raise TypeError(message)

    @property
    def hide(self) -> typing.Optional[bool]:
        r"""
        Is true when tuplet bracket hides.

        ..  container:: example

            >>> tuplet = abjad.Tuplet((2, 3), "c'8 d'8 e'8")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 2/3 {
                    c'8
                    d'8
                    e'8
                }

            >>> tuplet.hide is None
            True

        ..  container:: example

            >>> tuplet_1 = abjad.Tuplet((2, 3), "c'4 d'4 e'4")
            >>> tuplet_2 = abjad.Tuplet((2, 3), "d'4 e'4 f'4")
            >>> staff = abjad.Staff([tuplet_1, tuplet_2])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \times 2/3 {
                        c'4
                        d'4
                        e'4
                    }
                    \times 2/3 {
                        d'4
                        e'4
                        f'4
                    }
                }

            >>> staff[0].hide = True
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \scaleDurations #'(2 . 3) {
                        c'4
                        d'4
                        e'4
                    }
                    \times 2/3 {
                        d'4
                        e'4
                        f'4
                    }
                }

        Hides tuplet bracket and tuplet number when true.
        """
        return self._hide

    @hide.setter
    def hide(self, argument):
        assert isinstance(argument, (bool, type(None))), repr(argument)
        self._hide = argument

    @property
    def implied_prolation(self) -> Multiplier:
        r"""
        Gets implied prolation of tuplet.

        ..  container:: example

            Defined equal to tuplet multiplier:

            >>> tuplet = abjad.Tuplet((2, 3), "c'8 d'8 e'8")
            >>> abjad.show(tuplet) # doctest: +SKIP

            >>> tuplet.implied_prolation
            Multiplier(2, 3)

        """
        return self.multiplier

    @property
    def multiplied_duration(self) -> Duration:
        r"""
        Gets multiplied duration of tuplet.

        ..  container:: example

            >>> tuplet = abjad.Tuplet((2, 3), "c'8 d'8 e'8")

            >>> abjad.show(tuplet) # doctest: +SKIP

            >>> tuplet.multiplied_duration
            Duration(1, 4)

        """
        return self.multiplier * self._get_contents_duration()

    @property
    def multiplier(self) -> Multiplier:
        r"""
        Gets and sets multiplier of tuplet.

        ..  container:: example

            Gets tuplet multiplier:

                >>> tuplet = abjad.Tuplet((2, 3), "c'8 d'8 e'8")
                >>> abjad.show(tuplet) # doctest: +SKIP

            >>> tuplet.multiplier
            Multiplier(2, 3)

        ..  container:: example

            Sets tuplet multiplier:

                >>> tuplet.multiplier = abjad.Multiplier(4, 3)
                >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 4/3 {
                    c'8
                    d'8
                    e'8
                }

        """
        return self._multiplier

    @multiplier.setter
    def multiplier(self, argument):
        if isinstance(argument, (int, Fraction)):
            rational = Multiplier(argument)
        elif isinstance(argument, tuple):
            rational = Multiplier(argument)
        else:
            message = f'can not set tuplet multiplier: {argument!r}.'
            raise ValueError(message)
        if 0 < rational:
            self._multiplier = rational
        else:
            message = f'tuplet multiplier must be positive: {argument!r}.'
            raise ValueError(message)

    @property
    def tweaks(self):
        r"""
        Gets tweaks.

        ..  container:: example

            >>> tuplet_1 = abjad.Tuplet((2, 3), "c'4 ( d'4 e'4 )")
            >>> abjad.tweak(tuplet_1).color = 'red'
            >>> abjad.tweak(tuplet_1).staff_padding = 2

            >>> tuplet_2 = abjad.Tuplet((2, 3), "c'4 ( d'4 e'4 )")
            >>> abjad.tweak(tuplet_2).color = 'green'
            >>> abjad.tweak(tuplet_2).staff_padding = 2

            >>> tuplet_3 = abjad.Tuplet((5, 4), [tuplet_1, tuplet_2])
            >>> abjad.tweak(tuplet_3).color = 'blue'
            >>> abjad.tweak(tuplet_3).staff_padding = 4

            >>> staff = abjad.Staff([tuplet_3])
            >>> leaves = abjad.select(staff).leaves()
            >>> abjad.attach(abjad.TimeSignature((5, 4)), leaves[0]) 
            >>> literal = abjad.LilyPondLiteral(r'\set tupletFullLength = ##t')
            >>> abjad.attach(literal, staff)
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \set tupletFullLength = ##t
                    \tweak text #tuplet-number::calc-fraction-text
                    \tweak color #blue
                    \tweak staff-padding #4
                    \times 5/4 {
                        \tweak color #red
                        \tweak staff-padding #2
                        \times 2/3 {
                            \time 5/4
                            c'4
                            (
                            d'4
                            e'4
                            )
                        }
                        \tweak color #green
                        \tweak staff-padding #2
                        \times 2/3 {
                            c'4
                            (
                            d'4
                            e'4
                            )
                        }
                    }
                }

            ..  todo:: Report LilyPond bug that results from removing
                tupletFullLength in the example above: blue tuplet bracket
                shrinks to encompass only the second underlying tuplet.

        """
        return tweak(self)

    ### PUBLIC METHODS ###

    def append(self, component, preserve_duration=False) -> None:
        r"""
        Appends ``component`` to tuplet.

        ..  container:: example

            Appends note to tuplet:

            >>> tuplet = abjad.Tuplet((2, 3), "c'4 ( d'4 f'4 )")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 2/3 {
                    c'4
                    (
                    d'4
                    f'4
                    )
                }

            >>> tuplet.append(abjad.Note("e'4"))
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak edge-height #'(0.7 . 0)
                \times 2/3 {
                    c'4
                    (
                    d'4
                    f'4
                    )
                    e'4
                }

        ..  container:: example

            Appends note to tuplet and preserves tuplet duration:

            >>> tuplet = abjad.Tuplet((2, 3), "c'4 ( d'4 f'4 )")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 2/3 {
                    c'4
                    (
                    d'4
                    f'4
                    )
                }

            >>> tuplet.append(abjad.Note("e'4"), preserve_duration=True)
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 1/2 {
                    c'4
                    (
                    d'4
                    f'4
                    )
                    e'4
                }

        """
        if preserve_duration:
            old_duration = inspect(self).duration()
        Container.append(self, component)
        if preserve_duration:
            new_duration = self._get_contents_duration()
            multiplier = old_duration / new_duration
            self.multiplier = multiplier
            assert inspect(self).duration() == old_duration

    def augmentation(self) -> bool:
        r"""
        Is true when tuplet multiplier is greater than ``1``.

        ..  container:: example

            Augmented tuplet:

            >>> tuplet = abjad.Tuplet((4, 3), "c'8 d'8 e'8")
            >>> abjad.show(tuplet) # doctest: +SKIP

            >>> tuplet.augmentation()
            True

        ..  container:: example

            Diminished tuplet:

            >>> tuplet = abjad.Tuplet((2, 3), "c'4 d'4 e'4")
            >>> abjad.show(tuplet) # doctest: +SKIP

            >>> tuplet.augmentation()
            False

        ..  container:: example

            Trivial tuplet:

            >>> tuplet = abjad.Tuplet((1, 1), "c'8. d'8. e'8.")
            >>> abjad.show(tuplet) # doctest: +SKIP

            >>> tuplet.augmentation()
            False

        """
        if self.multiplier:
            return 1 < self.multiplier
        else:
            return False

    def diminution(self) -> bool:
        r"""
        Is true when tuplet multiplier is less than ``1``.

        ..  container:: example

            Augmented tuplet:

            >>> tuplet = abjad.Tuplet((4, 3), "c'8 d'8 e'8")
            >>> abjad.show(tuplet) # doctest: +SKIP

            >>> tuplet.diminution()
            False

        ..  container:: example

            Diminished tuplet:

            >>> tuplet = abjad.Tuplet((2, 3), "c'4 d'4 e'4")
            >>> abjad.show(tuplet) # doctest: +SKIP

            >>> tuplet.diminution()
            True

        ..  container:: example

            Trivial tuplet:

            >>> tuplet = abjad.Tuplet((1, 1), "c'8. d'8. e'8.")
            >>> abjad.show(tuplet) # doctest: +SKIP

            >>> tuplet.diminution()
            False

        """
        if self.multiplier:
            return self.multiplier < 1
        else:
            return False

    def extend(self, argument, preserve_duration=False) -> None:
        r"""
        Extends tuplet with ``argument``.

        ..  container:: example

            Extends tuplet with three notes:

            >>> tuplet = abjad.Tuplet((2, 3), "c'4 ( d'4 f'4 )")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 2/3 {
                    c'4
                    (
                    d'4
                    f'4
                    )
                }

            >>> notes = [abjad.Note("e'32"), abjad.Note("d'32"), abjad.Note("e'16")]
            >>> tuplet.extend(notes)
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak edge-height #'(0.7 . 0)
                \times 2/3 {
                    c'4
                    (
                    d'4
                    f'4
                    )
                    e'32
                    d'32
                    e'16
                }

        ..  container:: example

            Extends tuplet with three notes and preserves tuplet duration:

            >>> tuplet = abjad.Tuplet((2, 3), "c'4 ( d'4 f'4 )")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 2/3 {
                    c'4
                    (
                    d'4
                    f'4
                    )
                }

            >>> notes = [abjad.Note("e'32"), abjad.Note("d'32"), abjad.Note("e'16")]
            >>> tuplet.extend(notes, preserve_duration=True)
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 4/7 {
                    c'4
                    (
                    d'4
                    f'4
                    )
                    e'32
                    d'32
                    e'16
                }

        """
        if preserve_duration:
            old_duration = inspect(self).duration()
        Container.extend(self, argument)
        if preserve_duration:
            new_duration = self._get_contents_duration()
            multiplier = old_duration / new_duration
            self.multiplier = multiplier
            assert inspect(self).duration() == old_duration

    @staticmethod
    def from_duration(
        duration: typing.Union[typing.Tuple[int, int], Duration],
        components,
        ) -> 'Tuplet':
        r"""
        Makes tuplet from ``duration`` and ``components``.

        ..  container:: example

            Makes diminution:

            >>> tuplet = abjad.Tuplet.from_duration((2, 8), "c'8 d' e'")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 2/3 {
                    c'8
                    d'8
                    e'8
                }

        """
        if not len(components):
            raise Exception(f'components must be nonempty: {components!r}.')
        target_duration = Duration(duration)
        tuplet = Tuplet(1, components)
        contents_duration = inspect(tuplet).duration()
        multiplier = target_duration / contents_duration
        tuplet.multiplier = multiplier
        return tuplet

    @staticmethod
    def from_duration_and_ratio(
        duration,
        ratio,
        decrease_monotonic: bool = True,
        ) -> 'Tuplet':
        r"""
        Makes tuplet from ``duration`` and ``ratio``.

        ..  container:: example

            Makes tupletted leaves strictly without dots when all
            ``ratio`` equal ``1``:

            >>> tuplet = abjad.Tuplet.from_duration_and_ratio(
            ...     abjad.Duration(3, 16),
            ...     abjad.Ratio((1, 1, 1, -1, -1)),
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 3/16
                    \times 4/5 {
                        c'32.
                        c'32.
                        c'32.
                        r32.
                        r32.
                    }
                }   % measure

            Allows tupletted leaves to return with dots when some ``ratio``
            do not equal ``1``:

            >>> tuplet = abjad.Tuplet.from_duration_and_ratio(
            ...     abjad.Duration(3, 16),
            ...     abjad.Ratio((1, -2, -2, 3, 3)),
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 3/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 6/11 {
                        c'32
                        r16
                        r16
                        c'16.
                        c'16.
                    }
                }   % measure

            Interprets nonassignable ``ratio`` according to
            ``decrease_monotonic``:

            >>> tuplet = abjad.Tuplet.from_duration_and_ratio(
            ...     abjad.Duration(3, 16),
            ...     abjad.Ratio((5, -1, 5)),
            ...     decrease_monotonic=False,
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 3/16
                    \times 8/11 {
                        c'16...
                        r64.
                        c'16...
                    }
                }   % measure

        ..  container:: example

            Makes augmented tuplet from ``duration`` and ``ratio`` and
            encourages dots:

            >>> tuplet = abjad.Tuplet.from_duration_and_ratio(
            ...     abjad.Duration(3, 16),
            ...     abjad.Ratio((1, 1, 1, -1, -1)),
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 3/16
                    \times 4/5 {
                        c'32.
                        c'32.
                        c'32.
                        r32.
                        r32.
                    }
                }   % measure

            Interprets nonassignable ``ratio`` according to
            ``decrease_monotonic``:

            >>> tuplet = abjad.Tuplet.from_duration_and_ratio(
            ...     abjad.Duration(3, 16),
            ...     abjad.Ratio((5, -1, 5)),
            ...     decrease_monotonic=False,
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 3/16
                    \times 8/11 {
                        c'16...
                        r64.
                        c'16...
                    }
                }   % measure

        ..  container:: example

            Makes diminished tuplet from ``duration`` and nonzero integer
            ``ratio``.

            Makes tupletted leaves strictly without dots when all
            ``ratio`` equal ``1``:

            >>> tuplet = abjad.Tuplet.from_duration_and_ratio(
            ...     abjad.Duration(3, 16),
            ...     abjad.Ratio((1, 1, 1, -1, -1)),
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 3/16
                    \times 4/5 {
                        c'32.
                        c'32.
                        c'32.
                        r32.
                        r32.
                    }
                }   % measure

            Allows tupletted leaves to return with dots when some ``ratio``
            do not equal ``1``:

            >>> tuplet = abjad.Tuplet.from_duration_and_ratio(
            ...     abjad.Duration(3, 16),
            ...     abjad.Ratio((1, -2, -2, 3, 3)),
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 3/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 6/11 {
                        c'32
                        r16
                        r16
                        c'16.
                        c'16.
                    }
                }   % measure

            Interprets nonassignable ``ratio`` according to
            ``decrease_monotonic``:

            >>> tuplet = abjad.Tuplet.from_duration_and_ratio(
            ...     abjad.Duration(3, 16),
            ...     abjad.Ratio((5, -1, 5)),
            ...     decrease_monotonic=False,
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 3/16
                    \times 8/11 {
                        c'16...
                        r64.
                        c'16...
                    }
                }   % measure

        ..  container:: example

            Makes diminished tuplet from ``duration`` and ``ratio`` and
            encourages dots:

            >>> tuplet = abjad.Tuplet.from_duration_and_ratio(
            ...     abjad.Duration(3, 16),
            ...     abjad.Ratio((1, 1, 1, -1, -1)),
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type = 'RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 3/16
                    \times 4/5 {
                        c'32.
                        c'32.
                        c'32.
                        r32.
                        r32.
                    }
                }   % measure

            Interprets nonassignable ``ratio`` according to ``direction``:

            >>> tuplet = abjad.Tuplet.from_duration_and_ratio(
            ...     abjad.Duration(3, 16),
            ...     abjad.Ratio((5, -1, 5)),
            ...     decrease_monotonic=False,
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(measure) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 3/16
                    \times 8/11 {
                        c'16...
                        r64.
                        c'16...
                    }
                }   % measure

        Reduces ``ratio`` relative to each other.

        Interprets negative ``ratio`` as rests.

        Returns tuplet.
        """
        duration = Duration(duration)
        ratio = Ratio(ratio)
        basic_prolated_duration = duration / mathtools.weight(ratio.numbers)
        basic_written_duration = \
            basic_prolated_duration.equal_or_greater_assignable
        written_durations = [x * basic_written_duration for x in ratio.numbers]
        leaf_maker = LeafMaker(decrease_monotonic=decrease_monotonic)
        try:
            notes = [
                Note(0, x) if 0 < x else Rest(abs(x))
                for x in written_durations
                ]
        except exceptions.AssignabilityError:
            denominator = duration.denominator
            note_durations = [
                Duration(x, denominator)
                for x in ratio.numbers
                ]
            pitches = [
                None if note_duration < 0 else 0
                for note_duration in note_durations
                ]
            leaf_durations = [
                abs(note_duration)
                for note_duration in note_durations
                ]
            notes = leaf_maker(pitches, leaf_durations)
        tuplet = Tuplet.from_duration(duration, notes)
        tuplet.normalize_multiplier()
        return tuplet

    @staticmethod
    def from_leaf_and_ratio(
        leaf: Leaf,
        ratio: typing.Union[typing.List, Ratio],
        ) -> 'Tuplet':
        r"""
        Makes tuplet from ``leaf`` and ``ratio``.

        >>> note = abjad.Note("c'8.")

        ..  container:: example

            >>> tuplet = abjad.Tuplet.from_leaf_and_ratio(
            ...     note,
            ...     abjad.Ratio((1,)),
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 1/1 {
                    c'8.
                }

        ..  container:: example

            >>> tuplet = abjad.Tuplet.from_leaf_and_ratio(
            ...     note,
            ...     [1, 2],
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 1/1 {
                    c'16
                    c'8
                }

        ..  container:: example

            >>> tuplet = abjad.Tuplet.from_leaf_and_ratio(
            ...     note,
            ...     abjad.Ratio((1, 2, 2)),
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 4/5 {
                    c'32.
                    c'16.
                    c'16.
                }

        ..  container:: example

            >>> tuplet = abjad.Tuplet.from_leaf_and_ratio(
            ...     note,
            ...     [1, 2, 2, 3],
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 3/4 {
                    c'32
                    c'16
                    c'16
                    c'16.
                }

        ..  container:: example

            >>> tuplet = abjad.Tuplet.from_leaf_and_ratio(
            ...     note,
            ...     [1, 2, 2, 3, 3],
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 6/11 {
                    c'32
                    c'16
                    c'16
                    c'16.
                    c'16.
                }

        ..  container:: example

            >>> tuplet = abjad.Tuplet.from_leaf_and_ratio(
            ...     note,
            ...     abjad.Ratio((1, 2, 2, 3, 3, 4)),
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 4/5 {
                    c'64
                    c'32
                    c'32
                    c'32.
                    c'32.
                    c'16
                }

        ..  container:: example

            >>> tuplet = abjad.Tuplet.from_leaf_and_ratio(
            ...     note,
            ...     [1],
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 1/1 {
                    c'8.
                }

        ..  container:: example

            >>> tuplet = abjad.Tuplet.from_leaf_and_ratio(
            ...     note,
            ...     [1, 2],
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 1/1 {
                    c'16
                    c'8
                }

        ..  container:: example

            >>> tuplet = abjad.Tuplet.from_leaf_and_ratio(
            ...     note,
            ...     [1, 2, 2],
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 4/5 {
                    c'32.
                    c'16.
                    c'16.
                }

        ..  container:: example

            >>> tuplet = abjad.Tuplet.from_leaf_and_ratio(
            ...     note,
            ...     [1, 2, 2, 3],
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 3/4 {
                    c'32
                    c'16
                    c'16
                    c'16.
                }

        ..  container:: example

            >>> tuplet = abjad.Tuplet.from_leaf_and_ratio(
            ...     note,
            ...     [1, 2, 2, 3, 3],
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 6/11 {
                    c'32
                    c'16
                    c'16
                    c'16.
                    c'16.
                }

        ..  container:: example

            >>> tuplet = abjad.Tuplet.from_leaf_and_ratio(
            ...     note,
            ...     [1, 2, 2, 3, 3, 4],
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((3, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 4/5 {
                    c'64
                    c'32
                    c'32
                    c'32.
                    c'32.
                    c'16
                }

        Returns tuplet.
        """
        proportions = Ratio(ratio)
        target_duration = leaf.written_duration
        basic_prolated_duration = target_duration / sum(proportions.numbers)
        basic_written_duration = \
            basic_prolated_duration.equal_or_greater_assignable
        written_durations = [
            _ * basic_written_duration for _ in proportions.numbers
            ]
        maker = NoteMaker()
        try:
            notes = [Note(0, x) for x in written_durations]
        except exceptions.AssignabilityError:
            denominator = target_duration.denominator
            note_durations = [
                Duration(_, denominator)
                for _ in proportions.numbers
                ]
            notes = maker(0, note_durations)
        contents_duration = inspect(notes).duration()
        multiplier = target_duration / contents_duration
        tuplet = Tuplet(multiplier, notes)
        tuplet.normalize_multiplier()
        return tuplet

    @staticmethod
    def from_ratio_and_pair(
        ratio: typing.Union[typing.Tuple, NonreducedRatio],
        fraction: typing.Union[typing.Tuple, NonreducedFraction],
        ) -> 'Tuplet':
        r"""
        Makes tuplet from nonreduced ``ratio`` and nonreduced ``fraction``.

        ..  container:: example

            >>> tuplet = abjad.Tuplet.from_ratio_and_pair(
            ...     abjad.NonreducedRatio((1,)),
            ...     abjad.NonreducedFraction(7, 16),
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((7, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 7/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 1/1 {
                        c'4..
                    }
                }   % measure

            >>> tuplet = abjad.Tuplet.from_ratio_and_pair(
            ...     abjad.NonreducedRatio((1, 2)),
            ...     abjad.NonreducedFraction(7, 16),
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((7, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 7/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 7/6 {
                        c'8
                        c'4
                    }
                }   % measure

            >>> tuplet = abjad.Tuplet.from_ratio_and_pair(
            ...     abjad.NonreducedRatio((1, 2, 4)),
            ...     abjad.NonreducedFraction(7, 16),
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((7, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 7/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 1/1 {
                        c'16
                        c'8
                        c'4
                    }
                }   % measure

            >>> tuplet = abjad.Tuplet.from_ratio_and_pair(
            ...     abjad.NonreducedRatio((1, 2, 4, 1)),
            ...     abjad.NonreducedFraction(7, 16),
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((7, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 7/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 7/8 {
                        c'16
                        c'8
                        c'4
                        c'16
                    }
                }   % measure

            >>> tuplet = abjad.Tuplet.from_ratio_and_pair(
            ...     abjad.NonreducedRatio((1, 2, 4, 1, 2)),
            ...     abjad.NonreducedFraction(7, 16),
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((7, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 7/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 7/10 {
                        c'16
                        c'8
                        c'4
                        c'16
                        c'8
                    }
                }   % measure

            >>> tuplet = abjad.Tuplet.from_ratio_and_pair(
            ...     abjad.NonreducedRatio((1, 2, 4, 1, 2, 4)),
            ...     abjad.NonreducedFraction(7, 16),
            ...     )
            >>> staff = abjad.Staff(
            ...     [abjad.Measure((7, 16), [tuplet])],
            ...     lilypond_type='RhythmicStaff',
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff[0])
                {   % measure
                    \time 7/16
                    \times 1/2 {
                        c'16
                        c'8
                        c'4
                        c'16
                        c'8
                        c'4
                    }
                }   % measure

        Interprets ``d`` as tuplet denominator.
        """
        ratio = NonreducedRatio(ratio)
        if isinstance(fraction, tuple):
            fraction = NonreducedFraction(*fraction)
        numerator = fraction.numerator
        denominator = fraction.denominator
        duration = Duration(fraction)
        if len(ratio.numbers) == 1:
            if 0 < ratio.numbers[0]:
                try:
                    note = Note(0, duration)
                    duration = inspect(note).duration()
                    tuplet = Tuplet.from_duration(duration, [note])
                    return tuplet
                except exceptions.AssignabilityError:
                    note_maker = NoteMaker()
                    notes = note_maker(0, duration)
                    duration = inspect(notes).duration()
                    return Tuplet.from_duration(duration, notes)
            elif ratio.numbers[0] < 0:
                try:
                    rest = Rest(duration)
                    duration = inspect(rest).duration()
                    return Tuplet.from_duration(duration, [rest])
                except exceptions.AssignabilityError:
                    leaf_maker = LeafMaker()
                    rests = leaf_maker([None], duration)
                    duration = inspect(rests).duration()
                    return Tuplet.from_duration(duration, rests)
            else:
                raise ValueError('no divide zero values.')
        else:
            exponent = int(
                math.log(
                    mathtools.weight(ratio.numbers), 2) -
                    math.log(numerator, 2)
                    )
            denominator = int(denominator * 2 ** exponent)
            components = []
            for x in ratio.numbers:
                if not x:
                    raise ValueError('no divide zero values.')
                if 0 < x:
                    try:
                        note = Note(0, (x, denominator))
                        components.append(note)
                    except exceptions.AssignabilityError:
                        maker = NoteMaker()
                        notes = maker(0, (x, denominator))
                        components.extend(notes)
                else:
                    rests = Rest((-x, denominator))
                    components.append(rests)
            return Tuplet.from_duration(duration, components)

    def normalize_multiplier(self) -> None:
        r"""
        Normalizes tuplet multiplier.

        ..  container:: example

            >>> tuplet = abjad.Tuplet((1, 3), "c'4 d' e'")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 1/3 {
                    c'4
                    d'4
                    e'4
                }

            >>> tuplet.multiplier.normalized()
            False

            >>> tuplet.normalize_multiplier()
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 2/3 {
                    c'8
                    d'8
                    e'8
                }

            >>> tuplet.multiplier.normalized()
            True

        ..  container:: example

            >>> tuplet = abjad.Tuplet((8, 3), "c'32 d'32 e'32")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 8/3 {
                    c'32
                    d'32
                    e'32
                }

            >>> tuplet.multiplier.normalized()
            False

            >>> tuplet.normalize_multiplier()
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 4/3 {
                    c'16
                    d'16
                    e'16
                }

            >>> tuplet.multiplier.normalized()
            True

        ..  container:: example

            >>> tuplet = abjad.Tuplet((5, 12), "c'4 d'4 e'4")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 5/12 {
                    c'4
                    d'4
                    e'4
                }

            >>> tuplet.multiplier.normalized()
            False

            >>> tuplet.normalize_multiplier()
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 5/6 {
                    c'8
                    d'8
                    e'8
                }

            >>> tuplet.multiplier.normalized()
            True

        """
        # find tuplet multiplier
        integer_exponent = int(math.log(self.multiplier, 2))
        leaf_multiplier = Multiplier(2) ** integer_exponent
        # scale leaves in tuplet by power of two
        for component in self:
            if isinstance(component, Leaf):
                old_written_duration = component.written_duration
                new_written_duration = leaf_multiplier * old_written_duration
                component._set_duration(new_written_duration)
        numerator, denominator = leaf_multiplier.pair
        multiplier = Multiplier(denominator, numerator)
        self.multiplier *= multiplier

    def rewrite_dots(self) -> None:
        r"""
        Rewrites dots.

        ..  container:: example

            Rewrites single dots as 3:2 prolation:

            >>> tuplet = abjad.Tuplet(1, "c'8. c'8.")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 1/1 {
                    c'8.
                    c'8.
                }

            >>> tuplet.rewrite_dots()
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 3/2 {
                    c'8
                    c'8
                }

        ..  container:: example

            Rewrites double dots as 7:4 prolation:

            >>> tuplet = abjad.Tuplet(1, "c'8.. c'8..")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 1/1 {
                    c'8..
                    c'8..
                }

            >>> tuplet.rewrite_dots()
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 7/4 {
                    c'8
                    c'8
                }

        ..  container:: example

            Does nothing when dot counts differ:

            >>> tuplet = abjad.Tuplet(1, "c'8. d'8. e'8")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 1/1 {
                    c'8.
                    d'8.
                    e'8
                }

            >>> tuplet.rewrite_dots()
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 1/1 {
                    c'8.
                    d'8.
                    e'8
                }

        ..  container:: example

            Does nothing when leaves carry no dots:

            >>> tuplet = abjad.Tuplet((3, 2), "c'8 d' e'")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 3/2 {
                    c'8
                    d'8
                    e'8
                }

            >>> tuplet.rewrite_dots()
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 3/2 {
                    c'8
                    d'8
                    e'8
                }

        Not yet implemented for multiply nested tuplets.
        """
        dot_counts = set()
        for component in self:
            if isinstance(component, Tuplet):
                return
            dot_count = component.written_duration.dot_count
            dot_counts.add(dot_count)
        if 1 < len(dot_counts):
            return
        assert len(dot_counts) == 1
        global_dot_count = dot_counts.pop()
        if global_dot_count == 0:
            return
        dot_multiplier = Multiplier.from_dot_count(global_dot_count)
        self.multiplier *= dot_multiplier
        dot_multiplier_reciprocal = dot_multiplier.reciprocal
        for component in self:
            component.written_duration *= dot_multiplier_reciprocal

    def set_minimum_denominator(self, denominator) -> None:
        r"""
        Sets preferred denominator of tuplet to at least ``denominator``.

        ..  container:: example

            Sets preferred denominator of tuplet to ``8`` at least:

            >>> tuplet = abjad.Tuplet((3, 5), "c'4 d'8 e'8 f'4 g'2")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 3/5 {
                    c'4
                    d'8
                    e'8
                    f'4
                    g'2
                }

            >>> tuplet.set_minimum_denominator(8)
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 6/10 {
                    c'4
                    d'8
                    e'8
                    f'4
                    g'2
                }

        """
        assert mathtools.is_nonnegative_integer_power_of_two(denominator)
        self.force_fraction = True
        durations = [
            self._get_contents_duration(),
            self._get_preprolated_duration(),
            Duration(1, denominator),
            ]
        nonreduced_fractions = Duration.durations_to_nonreduced_fractions(
            durations)
        self.denominator = nonreduced_fractions[1].numerator

    def toggle_prolation(self) -> None:
        r"""
        Changes augmented tuplets to diminished;
        changes diminished tuplets to augmented.

        ..  container:: example

            Changes augmented tuplet to diminished:

            >>> tuplet = abjad.Tuplet((4, 3), "c'8 d'8 e'8")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 4/3 {
                    c'8
                    d'8
                    e'8
                }

            >>> tuplet.toggle_prolation()
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 2/3 {
                    c'4
                    d'4
                    e'4
                }

            Multiplies the written duration of the leaves in tuplet
            by the least power of ``2`` necessary to diminshed tuplet.

        ..  container:: example

            Changes diminished tuplet to augmented:

            >>> tuplet = abjad.Tuplet((2, 3), "c'4 d'4 e'4")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \times 2/3 {
                    c'4
                    d'4
                    e'4
                }

            >>> tuplet.toggle_prolation()
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 4/3 {
                    c'8
                    d'8
                    e'8
                }

            Divides the written duration of the leaves in tuplet
            by the least power of ``2`` necessary to diminshed tuplet.

        ..  container:: example

            REGRESSION. Leaves trivial tuplets unchanged:

            >>> tuplet = abjad.Tuplet(1, "c'4 d'4 e'4")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 1/1 {
                    c'4
                    d'4
                    e'4
                }

            >>> tuplet.toggle_prolation()
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 1/1 {
                    c'4
                    d'4
                    e'4
                }

        Does not yet work with nested tuplets.
        """
        if self.diminution():
            while self.diminution():
                self.multiplier *= 2
                for leaf in iterate(self).leaves():
                    leaf.written_duration /= 2
        elif self.augmentation():
            while not self.diminution():
                self.multiplier /= 2
                for leaf in iterate(self).leaves():
                    leaf.written_duration *= 2

    def trivial(self) -> bool:
        r"""
        Is true when tuplet multiplier is equal to ``1`` and no multipliers
        attach to any leaves in tuplet.

        ..  container:: example

            >>> tuplet = abjad.Tuplet((1, 1), "c'8 d'8 e'8")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 1/1 {
                    c'8
                    d'8
                    e'8
                }

            >>> tuplet.trivial()
            True

        ..  container:: example

            Tuplet is not trivial when multipliers attach to tuplet leaves:

            >>> tuplet = abjad.Tuplet((1, 1), "c'8 d'8 e'8")
            >>> abjad.attach(abjad.Multiplier(3, 2), tuplet[0])
            >>> abjad.attach(abjad.Multiplier(1, 2), tuplet[-1])
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 1/1 {
                    c'8 * 3/2
                    d'8
                    e'8 * 1/2
                }

            >>> tuplet.trivial()
            False

        """
        for leaf in iterate(self).leaves():
            if inspect(leaf).has_indicator(Multiplier):
                return False
        return self.multiplier == 1

    def trivializable(self) -> bool:
        r"""
        Is true when tuplet is trivializable (can be rewritten with a ratio of
        1:1).

        ..  container:: example

            Redudant tuplet:

            >>> tuplet = abjad.Tuplet((3, 4), "c'4 c'4")
            >>> measure = abjad.Measure((3, 8), [tuplet])
            >>> abjad.show(measure) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(measure)
                {   % measure
                    \time 3/8
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 3/4 {
                        c'4
                        c'4
                    }
                }   % measure

            >>> tuplet.trivializable()
            True

            Can be rewritten without a tuplet bracket:

                >>> measure = abjad.Measure((3, 8), "c'8. c'8.")
                >>> abjad.show(measure) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(measure)
                {   % measure
                    \time 3/8
                    c'8.
                    c'8.
                }   % measure

        ..  container:: example

            Nontrivializable tuplet:

            >>> tuplet = abjad.Tuplet((3, 5), "c'4 c'4 c'4 c'4 c'4")
            >>> measure = abjad.Measure((3, 4), [tuplet])
            >>> abjad.show(measure) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(measure)
                {   % measure
                    \time 3/4
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 3/5 {
                        c'4
                        c'4
                        c'4
                        c'4
                        c'4
                    }
                }   % measure

            >>> tuplet.trivializable()
            False

            Can not be rewritten without a tuplet bracket.

        ..  container:: example

            REGRESSION. Nontrivializable tuplet:

            >>> tuplet = abjad.Tuplet((3, 4), "c'2. c4")
            >>> measure = abjad.Measure((3, 4), [tuplet])
            >>> abjad.show(measure) # doctest: +SKIP

            ..  docs::

                        >>> abjad.f(measure)
                        {   % measure
                            \time 3/4
                            \tweak text #tuplet-number::calc-fraction-text
                            \times 3/4 {
                                c'2.
                                c4
                            }
                        }   % measure

            >>> tuplet.trivializable()
            False

        """
        for component in self:
            if isinstance(component, Tuplet):
                continue
            assert isinstance(component, Leaf), repr(component)
            duration = component.written_duration * self.multiplier
            if not duration.is_assignable:
                return False
        return True

    def trivialize(self) -> None:
        r"""
        Trivializes tuplet.

        ..  container:: example

            >>> tuplet = abjad.Tuplet((3, 4), "c'2")
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 3/4 {
                    c'2
                }

            >>> tuplet.trivializable()
            True

            >>> tuplet.trivialize()
            >>> abjad.show(tuplet) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(tuplet)
                \tweak text #tuplet-number::calc-fraction-text
                \times 1/1 {
                    c'4.
                }

        """
        if not self.trivializable():
            return
        for component in self:
            if isinstance(component, Tuplet):
                component.multiplier *= self.multiplier
            elif isinstance(component, Leaf):
                component.written_duration *= self.multiplier
            else:
                raise TypeError(component)
        self.multiplier = Multiplier(1)



import typing
from abjad import enums
from abjad import typings
from abjad.core.Chord import Chord
from abjad.core.Component import Component
from abjad.core.MultimeasureRest import MultimeasureRest
from abjad.core.Note import Note
from abjad.core.Rest import Rest
from abjad.core.Skip import Skip
from abjad.indicators.Clef import Clef
from abjad.top.detach import detach
from abjad.top.inspect import inspect
from abjad.top.iterate import iterate
from abjad.top.sequence import sequence
from abjad.top.tweak import tweak
from abjad.utilities.Duration import Duration
from abjad.utilities.DurationInequality import DurationInequality
from abjad.utilities.String import String
from .Spanner import Spanner


class Tie(Spanner):
    r"""
    Tie.

    ..  container:: example

        Ties four notes:

        >>> staff = abjad.Staff("c'4 c' c' c'")
        >>> abjad.attach(abjad.Tie(), staff[:])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                ~
                c'4
                ~
                c'4
                ~
                c'4
            }

    ..  container:: example

        Removes any existing ties before attaching new tie:

        >>> staff = abjad.Staff("c'4 ~ c' ~ c' ~ c'")
        >>> abjad.attach(abjad.Tie(), staff[:])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                ~
                c'4
                ~
                c'4
                ~
                c'4
            }

    ..  container:: example

        Ties consecutive chords if all adjacent pairs have at least one pitch
        in common:

        >>> staff = abjad.Staff("<c'>4 <c' d'>4 <d'>4")
        >>> abjad.attach(abjad.Tie(), staff[:])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                <c'>4
                ~
                <c' d'>4
                ~
                <d'>4
            }

    ..  container:: example

        Enharmonics are allowed:

        >>> staff = abjad.Staff("c'4 bs c' dff'")
        >>> abjad.attach(abjad.Tie(), staff[:])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                ~
                bs4
                ~
                c'4
                ~
                dff'4
            }

    ..  container:: example

        Raises exception when pitches do not equal each other:

        >>> staff = abjad.Staff("c'4 d' e'8 ~ e'8 r4")
        >>> abjad.attach(abjad.Tie(), staff[:2])
        Traceback (most recent call last):
            ...
        Exception: Tie()._attachment_test_all():
          Pitch {0} does not equal pitch {2}.

        Raises exception on nonnote, nonchord leaves:

        >>> staff = abjad.Staff("c'4 d' e'8 ~ e'8 r4")
        >>> abjad.attach(abjad.Tie(), staff[-2:])
        Traceback (most recent call last):
            ...
        Exception: Tie()._attachment_test_all():
          Can only tie notes and chords.
          Not Rest('r4').

        Detaches existing ties before attach:

        >>> staff = abjad.Staff("c'4 d' e'8 ~ e'8 r4")
        >>> abjad.attach(abjad.Tie(), staff[2:4])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                d'4
                e'8
                ~
                e'8
                r4
            }

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_direction',
        '_left_broken',
        '_repeat',
        '_right_broken',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        *,
        direction: typing.Union[str, enums.VerticalAlignment] = None,
        left_broken: bool = None,
        repeat: typing.Union[
            bool,
            typings.IntegerPair,
            DurationInequality,
            ] = None,
        right_broken: bool = None,
        ) -> None:
        Spanner.__init__(self)
        direction = String.to_tridirectional_lilypond_symbol(direction)
        self._direction = direction
        if left_broken is not None:
            left_broken = bool(left_broken)
        self._left_broken = left_broken
        repeat_ = repeat
        repeat_ = self._coerce_inequality(repeat)
        if repeat_ is not None:
            assert isinstance(repeat_, (bool, DurationInequality))
        self._repeat = repeat_
        if right_broken is not None:
            right_broken = bool(right_broken)
        self._right_broken = right_broken

    ### PRIVATE METHODS ###

    def _add_tweaks_and_direction(self, command):
        strings = []
        contributions = tweak(self)._list_format_contributions()
        strings.extend(contributions)
        string = self._add_direction(command)
        strings.append(string)
        return strings

    def _attachment_test(self, component):
        if self._ignore_attachment_test:
            return True
        if not isinstance(component, (Chord, Note)):
            return False
        if inspect(component).has_spanner(Tie):
            return False
        return True

    def _attachment_test_all(self, component_expression):
        if self._ignore_attachment_test:
            return True
        written_pitches = []
        if isinstance(component_expression, Component):
            component_expression = [component_expression]
        for component in component_expression:
            if isinstance(component, Note):
                written_pitches.append(set([component.written_pitch.number]))
            elif isinstance(component, Chord):
                numbers = [_.number for _ in component.written_pitches]
                written_pitches.append(set(numbers))
            else:
                return [
                    'Can only tie notes and chords.',
                    f'Not {component!r}.',
                    ]
        for pair in sequence(written_pitches).nwise():
            if not set.intersection(*pair):
                return [f'Pitch {pair[0]} does not equal pitch {pair[1]}.']
        return True

    def _before_attach(self, argument):
        if self._ignore_before_attach:
            return
        for leaf in iterate(argument).leaves():
            detach(Tie, leaf)

    def _can_have_conventional_tie(self, leaf):
        if self.repeat is True:
            return False
        if self.repeat in (False, None):
            return True
        assert isinstance(self.repeat, DurationInequality)
        return not self.repeat(leaf)

    def _can_have_repeat_tie(self, leaf):
        if self.repeat is None:
            return False
        if isinstance(self.repeat, bool):
            return self.repeat
        previous = inspect(leaf).leaf(-1)
        if previous is None:
            return True
        if previous not in self:
            return False
        assert isinstance(self.repeat, DurationInequality)
        return self.repeat(previous)

    @staticmethod
    def _coerce_inequality(argument):
        if isinstance(argument, tuple) and len(argument) == 2:
            return DurationInequality(
                operator_string='>=',
                duration=argument,
                )
        return argument

    def _conventional_tie_strings(self):
        return self._add_tweaks_and_direction('~')

    def _copy_keywords(self, new):
        new._direction = self.direction
        new._repeat = self.repeat

    def _get_lilypond_format_bundle(self, leaf):
        bundle = self._get_basic_lilypond_format_bundle(leaf)
        prototype = (
            MultimeasureRest,
            Rest,
            Skip,
            )
        if isinstance(leaf, prototype):
            return bundle
        if self._can_have_repeat_tie(leaf):
            if leaf is self[0]:
                if self._left_broken:
                    strings = self._repeat_tie_strings(leaf)
                    strings = self._tag_show(strings)
                    bundle.after.spanners.extend(strings)
            else:
                strings = self._repeat_tie_strings(leaf)
                bundle.after.spanners.extend(strings)
        if self._can_have_conventional_tie(leaf):
            if leaf is self[-1]:
                if self._right_broken:
                    strings = self._conventional_tie_strings()
                    strings = self._tag_show(strings)
                    bundle.after.spanners.extend(strings)
            elif isinstance(leaf._get_leaf(1), prototype):
                pass
            else:
                strings = self._conventional_tie_strings()
                bundle.after.spanners.extend(strings)
        return bundle

    def _repeat_tie_strings(self, leaf):
        strings = []
        if self._should_force_repeat_tie_up(leaf):
            strings.append(r'- \tweak direction #up')
        strings_ = self._add_tweaks_and_direction(r'\repeatTie')
        strings.extend(strings_)
        return strings

    @staticmethod
    def _should_force_repeat_tie_up(leaf):
        if not isinstance(leaf, (Note, Chord)):
            return False
        if leaf.written_duration < Duration(1):
            return False
        clef = inspect(leaf).effective(Clef, default=Clef('treble'))
        if isinstance(leaf, Note):
            written_pitches = [leaf.written_pitch]
        else:
            written_pitches = leaf.written_pitches
        for written_pitch in written_pitches:
            staff_position = written_pitch.to_staff_position(clef=clef)
            if staff_position.number == 0:
                return True
        return False

    ### PUBLIC PROPERTIES ###

    @property
    def cross_segment_examples(self):
        r"""
        Cross-segment examples.

        ..  container:: example

            [Tie] cross-segment example #1 (one-to-one):

            >>> segment_1 = abjad.Voice("c'4 d' f' f'", name='MainVoice')
            >>> tie = abjad.Tie(right_broken=True)
            >>> abjad.attach(tie, segment_1[-1:])
            >>> abjad.show(segment_1, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_1, strict=50)
                \context Voice = "MainVoice"
                {
                    c'4
                    d'4
                    f'4
                    f'4
                %@% ~                                             %! SHOW_TO_JOIN_BROKEN_SPANNERS
                }

            >>> segment_2 = abjad.Voice("f'4 f' d' c'", name='MainVoice')
            >>> abjad.show(segment_2, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_2, strict=50)
                \context Voice = "MainVoice"
                {
                    f'4
                    f'4
                    d'4
                    c'4
                }

            >>> container = abjad.Container([segment_1, segment_2])
            >>> text = format(container, 'lilypond')
            >>> text = abjad.LilyPondFormatManager.left_shift_tags(text, 50)
            >>> job = abjad.Job.join_broken_spanners(text)
            >>> text = job()
            >>> lines = text.split('\n')
            >>> lilypond_file = abjad.LilyPondFile.new(lines)
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> print(text)
                {
                    \context Voice = "MainVoice"
                    {
                        c'4
                        d'4
                        f'4
                        f'4
                        ~                                         %! SHOW_TO_JOIN_BROKEN_SPANNERS %@%
                    }
                    \context Voice = "MainVoice"
                    {
                        f'4
                        f'4
                        d'4
                        c'4
                    }
                }

        ..  container:: example

            [Tie] cross-segment example #2 (one-to-many):

            >>> segment_1 = abjad.Voice("c'4 d' f' f'", name='MainVoice')
            >>> tie = abjad.Tie(right_broken=True)
            >>> abjad.attach(tie, segment_1[-1:])
            >>> abjad.show(segment_1, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_1, strict=50)
                \context Voice = "MainVoice"
                {
                    c'4
                    d'4
                    f'4
                    f'4
                %@% ~                                             %! SHOW_TO_JOIN_BROKEN_SPANNERS
                }

            >>> segment_2 = abjad.Voice("f'4 f' d' c'", name='MainVoice')
            >>> abjad.attach(abjad.Tie(), segment_2[:2])
            >>> abjad.show(segment_2, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_2, strict=50)
                \context Voice = "MainVoice"
                {
                    f'4
                    ~
                    f'4
                    d'4
                    c'4
                }

            >>> container = abjad.Container([segment_1, segment_2])
            >>> text = format(container, 'lilypond')
            >>> text = abjad.LilyPondFormatManager.left_shift_tags(text, 50)
            >>> job = abjad.Job.join_broken_spanners(text)
            >>> text = job()
            >>> lines = text.split('\n')
            >>> lilypond_file = abjad.LilyPondFile.new(lines)
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> print(text)
                {
                    \context Voice = "MainVoice"
                    {
                        c'4
                        d'4
                        f'4
                        f'4
                        ~                                         %! SHOW_TO_JOIN_BROKEN_SPANNERS %@%
                    }
                    \context Voice = "MainVoice"
                    {
                        f'4
                        ~
                        f'4
                        d'4
                        c'4
                    }
                }

        ..  container:: example

            [Tie] cross-segment example #3 (many-to-one):

            >>> segment_1 = abjad.Voice("c'4 d' f' f'", name='MainVoice')
            >>> tie = abjad.Tie(right_broken=True)
            >>> abjad.attach(tie, segment_1[-2:])
            >>> abjad.show(segment_1, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_1, strict=50)
                \context Voice = "MainVoice"
                {
                    c'4
                    d'4
                    f'4
                    ~
                    f'4
                %@% ~                                             %! SHOW_TO_JOIN_BROKEN_SPANNERS
                }

            >>> segment_2 = abjad.Voice("f'4 f' d' c'", name='MainVoice')
            >>> abjad.show(segment_2, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_2, strict=50)
                \context Voice = "MainVoice"
                {
                    f'4
                    f'4
                    d'4
                    c'4
                }

            >>> container = abjad.Container([segment_1, segment_2])
            >>> text = format(container, 'lilypond')
            >>> text = abjad.LilyPondFormatManager.left_shift_tags(text, 50)
            >>> job = abjad.Job.join_broken_spanners(text)
            >>> text = job()
            >>> lines = text.split('\n')
            >>> lilypond_file = abjad.LilyPondFile.new(lines)
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> print(text)
                {
                    \context Voice = "MainVoice"
                    {
                        c'4
                        d'4
                        f'4
                        ~
                        f'4
                        ~                                         %! SHOW_TO_JOIN_BROKEN_SPANNERS %@%
                    }
                    \context Voice = "MainVoice"
                    {
                        f'4
                        f'4
                        d'4
                        c'4
                    }
                }

        ..  container:: example

            [Tie] cross-segment example #4 (many-to-many):

            >>> segment_1 = abjad.Voice("c'4 d' f' f'", name='MainVoice')
            >>> tie = abjad.Tie(right_broken=True)
            >>> abjad.attach(tie, segment_1[-2:])
            >>> abjad.show(segment_1, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_1, strict=50)
                \context Voice = "MainVoice"
                {
                    c'4
                    d'4
                    f'4
                    ~
                    f'4
                %@% ~                                             %! SHOW_TO_JOIN_BROKEN_SPANNERS
                }

            >>> segment_2 = abjad.Voice("f'4 f' d' c'", name='MainVoice')
            >>> abjad.attach(abjad.Tie(), segment_2[:2])

            ..  docs::

                >>> abjad.f(segment_2, strict=50)
                \context Voice = "MainVoice"
                {
                    f'4
                    ~
                    f'4
                    d'4
                    c'4
                }

            >>> container = abjad.Container([segment_1, segment_2])
            >>> text = format(container, 'lilypond')
            >>> text = abjad.LilyPondFormatManager.left_shift_tags(text, 50)
            >>> job = abjad.Job.join_broken_spanners(text)
            >>> text = job()
            >>> lines = text.split('\n')
            >>> lilypond_file = abjad.LilyPondFile.new(lines)
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> print(text)
                {
                    \context Voice = "MainVoice"
                    {
                        c'4
                        d'4
                        f'4
                        ~
                        f'4
                        ~                                         %! SHOW_TO_JOIN_BROKEN_SPANNERS %@%
                    }
                    \context Voice = "MainVoice"
                    {
                        f'4
                        ~
                        f'4
                        d'4
                        c'4
                    }
                }

        ..  container:: example

            [Repeat tie] cross-segment example #1 (one-to-one):

            >>> segment_1 = abjad.Voice("c'4 d' f' f'", name='MainVoice')
            >>> abjad.show(segment_1, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_1, strict=50)
                \context Voice = "MainVoice"
                {
                    c'4
                    d'4
                    f'4
                    f'4
                }

            >>> segment_2 = abjad.Voice("f'4 f' d' c'", name='MainVoice')
            >>> repeat_tie = abjad.Tie(repeat=True, left_broken=True)
            >>> abjad.attach(repeat_tie, segment_2[:1])
            >>> abjad.show(segment_2, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_2, strict=50)
                \context Voice = "MainVoice"
                {
                    f'4
                %@% \repeatTie                                    %! SHOW_TO_JOIN_BROKEN_SPANNERS
                    f'4
                    d'4
                    c'4
                }

            >>> container = abjad.Container([segment_1, segment_2])
            >>> text = format(container, 'lilypond')
            >>> text = abjad.LilyPondFormatManager.left_shift_tags(text, 50)
            >>> job = abjad.Job.join_broken_spanners(text)
            >>> text = job()
            >>> lines = text.split('\n')
            >>> lilypond_file = abjad.LilyPondFile.new(lines)
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> print(text)
                {
                    \context Voice = "MainVoice"
                    {
                        c'4
                        d'4
                        f'4
                        f'4
                    }
                    \context Voice = "MainVoice"
                    {
                        f'4
                        \repeatTie                                %! SHOW_TO_JOIN_BROKEN_SPANNERS %@%
                        f'4
                        d'4
                        c'4
                    }
                }

        ..  container:: example

            [Repeat tie] cross-segment example #2 (one-to-many):

            >>> segment_1 = abjad.Voice("c'4 d' f' f'", name='MainVoice')
            >>> abjad.show(segment_1, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_1, strict=50)
                \context Voice = "MainVoice"
                {
                    c'4
                    d'4
                    f'4
                    f'4
                }

            >>> segment_2 = abjad.Voice("f'4 f' d' c'", name='MainVoice')
            >>> repeat_tie = abjad.Tie(left_broken=True, repeat=True)
            >>> abjad.attach(repeat_tie, segment_2[:2])
            >>> abjad.show(segment_2, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_2, strict=50)
                \context Voice = "MainVoice"
                {
                    f'4
                %@% \repeatTie                                    %! SHOW_TO_JOIN_BROKEN_SPANNERS
                    f'4
                    \repeatTie
                    d'4
                    c'4
                }

            >>> container = abjad.Container([segment_1, segment_2])
            >>> text = format(container, 'lilypond')
            >>> text = abjad.LilyPondFormatManager.left_shift_tags(text, 50)
            >>> job = abjad.Job.join_broken_spanners(text)
            >>> text = job()
            >>> lines = text.split('\n')
            >>> lilypond_file = abjad.LilyPondFile.new(lines)
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> print(text)
                {
                    \context Voice = "MainVoice"
                    {
                        c'4
                        d'4
                        f'4
                        f'4
                    }
                    \context Voice = "MainVoice"
                    {
                        f'4
                        \repeatTie                                %! SHOW_TO_JOIN_BROKEN_SPANNERS %@%
                        f'4
                        \repeatTie
                        d'4
                        c'4
                    }
                }

        ..  container:: example

            [Repeat tie] cross-segment example #3 (many-to-one):

            >>> segment_1 = abjad.Voice("c'4 d' f' f'", name='MainVoice')
            >>> repeat_tie = abjad.Tie(repeat=True)
            >>> abjad.attach(repeat_tie, segment_1[-2:])
            >>> abjad.show(segment_1, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_1, strict=50)
                \context Voice = "MainVoice"
                {
                    c'4
                    d'4
                    f'4
                    f'4
                    \repeatTie
                }

            >>> segment_2 = abjad.Voice("f'4 f' d' c'", name='MainVoice')
            >>> repeat_tie = abjad.Tie(left_broken=True, repeat=True)
            >>> abjad.attach(repeat_tie, segment_2[:1])
            >>> abjad.show(segment_2, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_2, strict=50)
                \context Voice = "MainVoice"
                {
                    f'4
                %@% \repeatTie                                    %! SHOW_TO_JOIN_BROKEN_SPANNERS
                    f'4
                    d'4
                    c'4
                }

            >>> container = abjad.Container([segment_1, segment_2])
            >>> text = format(container, 'lilypond')
            >>> text = abjad.LilyPondFormatManager.left_shift_tags(text, 50)
            >>> job = abjad.Job.join_broken_spanners(text)
            >>> text = job()
            >>> lines = text.split('\n')
            >>> lilypond_file = abjad.LilyPondFile.new(lines)
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> print(text)
                {
                    \context Voice = "MainVoice"
                    {
                        c'4
                        d'4
                        f'4
                        f'4
                        \repeatTie
                    }
                    \context Voice = "MainVoice"
                    {
                        f'4
                        \repeatTie                                %! SHOW_TO_JOIN_BROKEN_SPANNERS %@%
                        f'4
                        d'4
                        c'4
                    }
                }

        ..  container:: example

            [Repeat tie] cross-segment example #4 (many-to-many):

            >>> segment_1 = abjad.Voice("c'4 d' f' f'", name='MainVoice')
            >>> repeat_tie = abjad.Tie(repeat=True)
            >>> abjad.attach(repeat_tie, segment_1[-2:])
            >>> abjad.show(segment_1, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_1, strict=50)
                \context Voice = "MainVoice"
                {
                    c'4
                    d'4
                    f'4
                    f'4
                    \repeatTie
                }

            >>> segment_2 = abjad.Voice("f'4 f' d' c'", name='MainVoice')
            >>> repeat_tie = abjad.Tie(left_broken=True, repeat=True)
            >>> abjad.attach(repeat_tie, segment_2[:2])

            ..  docs::

                >>> abjad.f(segment_2, strict=50)
                \context Voice = "MainVoice"
                {
                    f'4
                %@% \repeatTie                                    %! SHOW_TO_JOIN_BROKEN_SPANNERS
                    f'4
                    \repeatTie
                    d'4
                    c'4
                }

            >>> container = abjad.Container([segment_1, segment_2])
            >>> text = format(container, 'lilypond')
            >>> text = abjad.LilyPondFormatManager.left_shift_tags(text, 50)
            >>> job = abjad.Job.join_broken_spanners(text)
            >>> text = job()
            >>> lines = text.split('\n')
            >>> lilypond_file = abjad.LilyPondFile.new(lines)
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> print(text)
                {
                    \context Voice = "MainVoice"
                    {
                        c'4
                        d'4
                        f'4
                        f'4
                        \repeatTie
                    }
                    \context Voice = "MainVoice"
                    {
                        f'4
                        \repeatTie                                %! SHOW_TO_JOIN_BROKEN_SPANNERS %@%
                        f'4
                        \repeatTie
                        d'4
                        c'4
                    }
                }

        """
        pass

    @property
    def direction(self) -> typing.Optional[String]:
        r"""
        Gets direction.

        ..  container:: example

            Forces ties up:

            >>> staff = abjad.Staff("c'8 c'8 c'8 c'8")
            >>> tie = abjad.Tie(direction=abjad.Up)
            >>> abjad.attach(tie, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    ^ ~
                    c'8
                    ^ ~
                    c'8
                    ^ ~
                    c'8
                }

            >>> tie.direction
            '^'

        ..  container:: example

            Forces ties down:

            >>> staff = abjad.Staff("c'8 c'8 c'8 c'8")
            >>> tie = abjad.Tie(direction=abjad.Down)
            >>> abjad.attach(tie, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    _ ~
                    c'8
                    _ ~
                    c'8
                    _ ~
                    c'8
                }

            >>> tie.direction
            '_'

        ..  container:: example

            Positions ties according to LilyPond defaults:

            >>> staff = abjad.Staff("c'8 c'8 c'8 c'8")
            >>> tie = abjad.Tie(direction=None)
            >>> abjad.attach(tie, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    ~
                    c'8
                    ~
                    c'8
                    ~
                    c'8
                }

            >>> tie.direction is None
            True

        """
        return self._direction

    @property
    def left_broken(self) -> typing.Optional[bool]:
        """
        Is true when spanner is left-broken.
        """
        return self._left_broken

    @property
    def repeat(self) -> typing.Union[
        bool, DurationInequality, typings.IntegerPair, None,
        ]:
        r"""
        Gets repeat-tie threshold.

        ..  container:: example

            Formats all ties as repeat-ties when ``repeat`` is true:

            >>> staff = abjad.Staff("c'2 c'8 c'4.")
            >>> tie = abjad.Tie(repeat=True)
            >>> abjad.attach(tie, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'2
                    c'8
                    \repeatTie
                    c'4.
                    \repeatTie
                }

        ..  container:: example

            Formats only ties that satisfy duration inequality when ``repeat``
            is a duration inequality:

            >>> staff = abjad.Staff("c'2 c'8 c'4.")
            >>> repeat = abjad.DurationInequality(
            ...     operator_string='>=',
            ...     duration=abjad.Duration(1, 4),
            ...     )
            >>> tie = abjad.Tie(repeat=repeat)
            >>> abjad.attach(tie, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..   docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'2
                    c'8
                    \repeatTie
                    ~
                    c'4.
                }

            Coerces integer pair to >= inequality:

            >>> staff = abjad.Staff("c'2 c'8 c'4.")
            >>> tie = abjad.Tie(repeat=(1, 4))
            >>> abjad.attach(tie, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..   docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'2
                    c'8
                    \repeatTie
                    ~
                    c'4.
                }

            Durations that satisfy inequality can be said to "meet repeat-tie
            threshold." Durations that do not meet repeat-tie threshold format
            conventional tie on current note; durations that do meet repeat-tie
            threshold format repeat-tie on following note.

        ..  container:: example

            LILYPOND FIX. Automatically tweaks repeat tie direction up when
            repeat tie connects to long-duration note at staff position zero:

            >>> tie = abjad.Tie(repeat=True)
            >>> staff = abjad.Staff(r"b'4 b'4 b'2 b'1 b'\breve")
            >>> abjad.attach(abjad.TimeSignature((8, 4)), staff[-1])
            >>> abjad.attach(tie, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..   docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    b'4
                    b'4
                    \repeatTie
                    b'2
                    \repeatTie
                    b'1
                    - \tweak direction #up
                    \repeatTie
                    \time 8/4
                    b'\breve
                    - \tweak direction #up
                    \repeatTie
                }

            Without this fix, LilyPond incorrectly down-renders the last two
            repeat-ties in the example above.

        """
        return self._repeat

    @property
    def right_broken(self) -> typing.Optional[bool]:
        """
        Is true when spanner is right-broken.
        """
        return self._right_broken



from abjad.system.AbjadObject import AbjadObject


class LilyPondFraction(AbjadObject):
    """
    Model of a fraction in LilyPond.

    Not composer-safe.

    Used internally by LilyPondParser.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        'numerator',
        'denominator',
        )

    ### INITIALIZER ###

    def __init__(self, numerator=0, denominator=1):
        self.numerator = numerator
        self.denominator = denominator



import collections
from abjad import Fraction


def arithmetic_mean(argument):
    """
    Gets arithmetic mean of ``argument``.

    ..  container:: example

        >>> abjad.mathtools.arithmetic_mean([1, 2, 2, 20, 30])
        11

        >>> abjad.mathtools.arithmetic_mean([1, 2, 20])
        Fraction(23, 3)

        >>> abjad.mathtools.arithmetic_mean([2, 2, 20.0])
        8.0

    Raises exception when ``argument`` is not iterable.

    Returns number.
    """
    if not isinstance(argument, collections.Iterable):
        raise TypeError(argument)
    total = sum(argument)
    length = len(argument)
    if isinstance(total, float):
        return total / length
    result = Fraction(sum(argument), len(argument))
    int_result = int(result)
    if int_result == result:
        return int_result
    else:
        return result



#! /usr/bin/env python
import abjad
import baca
import ide
import os
import pathlib
import sys
import traceback


if __name__ == '__main__':

    layout_module_name = '{layout_module_name}'
    maker = ide.Path(os.path.realpath(__file__))

    try:
        from {layout_module_name} import breaks
        assert isinstance(breaks, baca.BreakMeasureMap), repr(breaks)
    except ImportError:
        print(f'No breaks in {{maker.trim()}} ...')
        sys.exit(1)

    try:
        from {layout_module_name} import spacing
        prototype = baca.HorizontalSpacingSpecifier
        assert isinstance(spacing, prototype), repr(spacing)
    except ImportError:
        spacing = None

    try:
        buildspace_directory = maker.parent
        layout_py = buildspace_directory('{layout_module_name}.py')
        document_name = abjad.String(buildspace_directory.name).to_shout_case()
    except:
        traceback.print_exc()
        sys.exit(1)

    try:
        if buildspace_directory.get_metadatum('parts_directory') is True:
            from {layout_module_name} import part_identifier
            assert abjad.String(part_identifier).is_shout_case()
            document_name = f'{{document_name}}_{{part_identifier}}'
    except ImportError:
        traceback.print_exc()
        sys.exit(1)

    try:
        if buildspace_directory.is_segment():
            string = 'first_measure_number'
            first_measure_number = buildspace_directory.get_metadatum(string)
            if not bool(first_measure_number):
                print('Can not find first measure number ...')
                first_measure_number = False
            assert isinstance(first_measure_number, int)
        else:
            first_measure_number = 1
    except:
        traceback.print_exc()
        sys.exit(1)

    if first_measure_number is False:
        print('Skipping layout ...')
        sys.exit(1)
        
    try:
        assert abjad.String(document_name).is_shout_case()
        string = 'first_measure_number'
        first_measure_number = buildspace_directory.get_metadatum(string, 1)
        time_signatures = buildspace_directory.get_time_signature_metadata()
        if breaks.partial_score is not None:
            time_signatures = time_signatures[:breaks.partial_score]
    except:
        traceback.print_exc()
        sys.exit(1)

    try:
        print(' Running segment-maker ...')
        maker = baca.SegmentMaker(
            breaks=breaks,
            do_not_check_persistence=True,
            do_not_include_layout_ly=True,
            final_bar_line=False,
            first_measure_number=first_measure_number,
            score_template=baca.SingleStaffScoreTemplate(),
            spacing=spacing,
            time_signatures=time_signatures,
            )
        lilypond_file = maker.run(remove=abjad.tags.layout_removal_tags())
        context = lilypond_file['GlobalSkips']
        skips = baca.select(context).skips()
        for skip in skips:
            abjad.detach(abjad.TimeSignature, skip)
        score = lilypond_file['Score']
        del(score['MusicContext'])
    except:
        traceback.print_exc()
        sys.exit(1)

    try:
        score = lilypond_file['Score']
        text = format(score, 'lilypond')
        text = text.replace('GlobalSkips', 'PageLayout')
        text = abjad.LilyPondFormatManager.left_shift_tags(text, realign=89)
        time_signatures = [str(_) for _ in time_signatures]
        line_1 = f'% time_signatures = {{time_signatures}}\n'
        measure_count = len(time_signatures)
        line_2 = f'% measure_count = {{measure_count}}\n'
        if breaks.partial_score is not None:
            line_3 = f'% partial_score = True'
            text = line_1 + line_2 + line_3 + '\n\n' + text
        else:
            text = line_1 + line_2 + '\n\n' + text
        layout_ly = layout_module_name.replace('_', '-') + '.ly'
        layout_ly = buildspace_directory(layout_ly)
        layout_ly.write_text(text)
        counter = abjad.String('measure').pluralize(measure_count)
        message = f' Writing {{measure_count}} {{counter}}'
        message += f' to {{layout_ly.trim()}} ...'
        print(message)
    except:
        traceback.print_exc()
        sys.exit(1)

    try:
        bol_measure_numbers = []
        prototype = abjad.LilyPondLiteral
        skips = abjad.iterate(score['GlobalSkips']).leaves(abjad.Skip)
        for i, skip in enumerate(skips):
            for literal in abjad.inspect(skip).indicators(prototype):
                if literal.argument in (r'\break', r'\pageBreak'):
                    measure_number = first_measure_number + i
                    bol_measure_numbers.append(measure_number)
                    continue
        bols = bol_measure_numbers
        count = len(bols)
        numbers = abjad.String('number').pluralize(count)
        if count <= 4:
            items = ', '.join([str(_) for _ in bols])
            print(
                f' Writing BOL measure {{numbers}} {{items}} to metadata ...')
        else:
            print(f' Writing BOL measure {{numbers}} to metadata ...')
            parts = abjad.sequence(bols).partition_by_counts(
                [12],
                cyclic=True,
                overhang=True,
                )
            for part in parts:
                items = ', '.join(str(_) for _ in part)
                print(f'  {{items}} ...')
        buildspace_directory.add_buildspace_metadatum(
            'bol_measure_numbers',
            bol_measure_numbers,
            document_name=document_name,
            )
    except:
        traceback.print_exc()
        sys.exit(1)

    sys.exit(0)



from abjad.utilities.OrderedDict import OrderedDict
from .ScoreTemplate import ScoreTemplate


class TwoStaffPianoScoreTemplate(ScoreTemplate):
    """
    Two-staff piano score template.

    ..  container:: example

        >>> template = abjad.TwoStaffPianoScoreTemplate()
        >>> abjad.show(template) # doctest: +SKIP

        >>> abjad.f(template.__illustrate__()[abjad.Score])
        \context Score = "TwoStaffPianoScore"
        <<
            \context GlobalContext = "GlobalContext"
            <<
                \context GlobalRests = "GlobalRests"
                {
                }
                \context GlobalSkips = "GlobalSkips"
                {
                }
            >>
            \context PianoStaff = "PianoStaff"
            <<
                \context Staff = "RHStaff"
                {
                    \context Voice = "RHVoice"
                    {
                        s1
                    }
                }
                \context Staff = "LHStaff"
                {
                    \context Voice = "LHVoice"
                    {
                        \clef "bass" %! ST3
                        s1
                    }
                }
            >>
        >>

    Returns score template.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        )

    ### INITIALIZER ###

    def __init__(self):
        super().__init__()
        self.voice_abbreviations.update({
            'rh': 'RHVoice',
            'lh': 'LHVoice',
            })

    ### SPECIAL METHODS ###

    def __call__(self):
        """
        Calls two-staff piano score template.

        Returns score.
        """
        import abjad
        # GLOBAL CONTEXT
        global_context = self._make_global_context()

        # RH STAFF
        rh_voice = abjad.Voice(name='RHVoice')
        rh_staff = abjad.Staff(
            [rh_voice],
            name='RHStaff',
            )

        # LH STAFF
        lh_voice = abjad.Voice(name='LHVoice')
        lh_staff = abjad.Staff(
            [lh_voice],
            name='LHStaff',
            )
        abjad.annotate(
            lh_staff,
            'default_clef',
            abjad.Clef('bass'),
            )

        # PIANO STAFF
        staff_group = abjad.StaffGroup(
            [rh_staff, lh_staff],
            lilypond_type='PianoStaff',
            name='PianoStaff',
            )
        abjad.annotate(
            staff_group,
            'default_instrument',
            abjad.Piano(),
            )

        # SCORE
        score = abjad.Score(
            [global_context, staff_group],
            name='TwoStaffPianoScore',
            )
        return score



from abjad import mathtools
from abjad.top.new import new
from .Segment import Segment


class IntervalClassSegment(Segment):
    """
    Interval-class segment.

    ..  container:: example

        An interval-class segment:

        >>> intervals = 'm2 M10 -aug4 P5'
        >>> abjad.IntervalClassSegment(intervals)
        IntervalClassSegment(['+m2', '+M3', '-A4', '+P5'])

    ..  container:: example

        Another interval-class segment:

        >>> intervals = 'P4 P5 P11 P12'
        >>> abjad.IntervalClassSegment(intervals)
        IntervalClassSegment(['+P4', '+P5', '+P4', '+P5'])

    Returns interval-class segment.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### PRIVATE PROPERTIES ###

    @property
    def _named_item_class(self):
        import abjad
        return abjad.NamedIntervalClass

    @property
    def _numbered_item_class(self):
        import abjad
        return abjad.NumberedIntervalClass

    @property
    def _parent_item_class(self):
        import abjad
        return abjad.IntervalClass

    ### PUBLIC PROPERTIES ###

    @property
    def is_tertian(self):
        """
        Is true when all named interval-classes in segment are tertian.

        ..  container:: example

            >>> interval_class_segment = abjad.IntervalClassSegment(
            ...     items=[('major', 3), ('minor', 6), ('major', 6)],
            ...     item_class=abjad.NamedIntervalClass,
            ...     )
            >>> interval_class_segment.is_tertian
            True

        Returns true or false.
        """
        import abjad
        inversion_equivalent_interval_class_segment = new(
            self,
            item_class=abjad.NamedInversionEquivalentIntervalClass,
            )
        for interval in inversion_equivalent_interval_class_segment:
            if not interval.number == 3:
                return False
        return True

    ### PUBLIC METHODS ###

    @classmethod
    def from_selection(class_, selection, item_class=None):
        """
        Initializes interval-class segment from component selection.

        ..  container:: example

            >>> staff_1 = abjad.Staff("c'4 <d' fs' a'>4 b2")
            >>> staff_2 = abjad.Staff("c4. r8 g2")
            >>> selection = abjad.select((staff_1, staff_2))
            >>> abjad.IntervalClassSegment.from_selection(selection)
            IntervalClassSegment(['-M2', '-M3', '-m3', '+m7', '+M7', '-P5'])

        Returns interval-class segment.
        """
        import abjad
        pitch_segment = abjad.PitchSegment.from_selection(selection)
        pitches = [_ for _ in pitch_segment]
        intervals = mathtools.difference_series(pitches)
        return class_(
            items=intervals,
            item_class=item_class,
            )

    def has_duplicates(self):
        """
        Is true when segment contains duplicates.

        ..  container:: example

            >>> intervals = 'm2 M3 -aug4 m2 P5'
            >>> segment = abjad.IntervalClassSegment(intervals)
            >>> segment.has_duplicates()
            True

            >>> intervals = 'M3 -aug4 m2 P5'
            >>> segment = abjad.IntervalClassSegment(intervals)
            >>> segment.has_duplicates()
            False

        Returns true or false.
        """
        import abjad
        return len(abjad.IntervalClassSet(
            self, item_class=self.item_class)) < len(self)



import functools
import numbers
from abjad import enums
from abjad import mathtools
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.system.FormatSpecification import FormatSpecification
from . import constants


@functools.total_ordering
class Accidental(AbjadValueObject):
    """
    Accidental.

    ..  container:: example

        >>> abjad.Accidental('ff')
        Accidental('double flat')

        >>> abjad.Accidental('tqf')
        Accidental('three-quarters flat')

        >>> abjad.Accidental('f')
        Accidental('flat')

        >>> abjad.Accidental('')
        Accidental('natural')

        >>> abjad.Accidental('qs')
        Accidental('quarter sharp')

        >>> abjad.Accidental('s')
        Accidental('sharp')

        >>> abjad.Accidental('tqs')
        Accidental('three-quarters sharp')

        >>> abjad.Accidental('ss')
        Accidental('double sharp')

    ..  container:: example

        Generalized accidentals are allowed:

        >>> abjad.Accidental('ssss')
        Accidental('ssss')

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_arrow',
        '_semitones',
        )

    ### INITIALIZER ##

    def __init__(self, name='', *, arrow=None):
        import abjad
        semitones = 0
        _arrow = None
        if name is None:
            pass
        elif isinstance(name, str):
            if name in constants._accidental_name_to_abbreviation:
                name = constants._accidental_name_to_abbreviation[name]
                semitones = constants._accidental_abbreviation_to_semitones[name]
            else:
                match = constants._comprehensive_accidental_regex.match(name)
                if not match:
                    try:
                        pitch = abjad.NamedPitch(name)
                        semitones = pitch.accidental.semitones
                        _arrow = pitch.accidental.arrow
                    except Exception:
                        message = 'can not instantiate {} from {!r}.'
                        message = message.format(type(self).__name__, name)
                        raise TypeError(message)
                else:
                    group_dict = match.groupdict()
                    if group_dict['alphabetic_accidental']:
                        prefix, _, suffix = name.partition('q')
                        if prefix.startswith('s'):
                            semitones += len(prefix)
                        elif prefix.startswith('f'):
                            semitones -= len(prefix)
                        if suffix == 's':
                            semitones += 0.5
                            if prefix == 't':
                                semitones += 1
                        elif suffix == 'f':
                            semitones -= 0.5
                            if prefix == 't':
                                semitones -= 1
                    elif group_dict['symbolic_accidental']:
                        semitones += name.count('#')
                        semitones -= name.count('b')
                        if name.endswith('+'):
                            semitones += 0.5
                        elif name.endswith('~'):
                            semitones -= 0.5
        elif isinstance(name, numbers.Number):
            semitones = float(name)
            assert (semitones % 1.) in (0., 0.5)
        elif hasattr(name, 'accidental'):
            _arrow = name.accidental.arrow
            semitones = name.accidental.semitones
        elif isinstance(name, type(self)):
            _arrow = name.arrow
            semitones = name.semitones
        else:
            try:
                pitch = abjad.NamedPitch(name)
                semitones = pitch.accidental.semitones
                _arrow = pitch.accidental.arrow
            except Exception:
                message = 'can not initialize accidental from value: {!r}'
                message = message.format(name)
                raise ValueError(message)
        semitones = mathtools.integer_equivalent_number_to_integer(semitones)
        self._semitones = semitones
        self._arrow = _arrow
        if arrow is not None:
            arrow = enums.VerticalAlignment.from_expr(arrow)
            if arrow is enums.Center:
                arrow = None
            self._arrow = arrow

    ### SPECIAL METHODS ###

    def __add__(self, argument):
        """
        Adds `argument` to accidental.

        ..  container:: example

            >>> accidental = abjad.Accidental('qs')

            >>> accidental + accidental
            Accidental('sharp')

            >>> accidental + accidental + accidental
            Accidental('three-quarters sharp')

        Returns new accidental.
        """
        if not isinstance(argument, type(self)):
            message = 'can only add accidental to other accidental.'
            raise TypeError(message)
        semitones = self.semitones + argument.semitones
        return type(self)(semitones)

    def __call__(self, argument):
        """
        Calls accidental on `argument`.

        >>> accidental = abjad.Accidental('s')

        ..  container:: example

            Calls accidental on pitches:

            >>> accidental(abjad.NamedPitch("c''"))
            NamedPitch("cs''")

            >>> accidental(abjad.NamedPitch("cqs''"))
            NamedPitch("ctqs''")

            >>> accidental(abjad.NumberedPitch(12))
            NumberedPitch(13)

            >>> accidental(abjad.NumberedPitch(12.5))
            NumberedPitch(13.5)

        ..  container:: example

            Calls accidental on pitch-classes:

            >>> accidental(abjad.NamedPitchClass('c'))
            NamedPitchClass('cs')

            >>> accidental(abjad.NamedPitchClass('cqs'))
            NamedPitchClass('ctqs')

            >>> accidental(abjad.NumberedPitchClass(0))
            NumberedPitchClass(1)

            >>> accidental(abjad.NumberedPitchClass(0.5))
            NumberedPitchClass(1.5)

        Returns new object of `argument` type.
        """
        if not hasattr(argument, '_apply_accidental'):
            message = 'do not know how to apply accidental to {!r}.'
            message = message.format(argument)
            raise TypeError(message)
        return argument._apply_accidental(self)

    def __lt__(self, argument):
        """
        Is true when `argument` is an accidental with semitones greater
        than those of this accidental.

        ..  container:: example

            >>> accidental_1 = abjad.Accidental('f')
            >>> accidental_2 = abjad.Accidental('f')
            >>> accidental_3 = abjad.Accidental('s')

            >>> accidental_1 < accidental_1
            False
            >>> accidental_1 < accidental_2
            False
            >>> accidental_1 < accidental_3
            True

            >>> accidental_2 < accidental_1
            False
            >>> accidental_2 < accidental_2
            False
            >>> accidental_2 < accidental_3
            True

            >>> accidental_3 < accidental_1
            False
            >>> accidental_3 < accidental_2
            False
            >>> accidental_3 < accidental_3
            False

        Returns true or false.
        """
        return self.semitones < argument.semitones

    def __neg__(self):
        """
        Negates accidental.

        ..  container:: example

            >>> -abjad.Accidental('ff')
            Accidental('double sharp')

            >>> -abjad.Accidental('tqf')
            Accidental('three-quarters sharp')

            >>> -abjad.Accidental('f')
            Accidental('sharp')

            >>> -abjad.Accidental('')
            Accidental('natural')

            >>> -abjad.Accidental('qs')
            Accidental('quarter flat')

            >>> -abjad.Accidental('s')
            Accidental('flat')

            >>> -abjad.Accidental('tqs')
            Accidental('three-quarters flat')

            >>> -abjad.Accidental('ss')
            Accidental('double flat')

        Returns new accidental.
        """
        return type(self)(-self.semitones)

    def __radd__(self, argument):
        """
        Raises not implemented error on accidental.
        """
        raise NotImplementedError

    def __str__(self):
        """
        Gets string representation of accidental.

        ..  container:: example

            >>> str(abjad.Accidental('ff'))
            'ff'

            >>> str(abjad.Accidental('tqf'))
            'tqf'

            >>> str(abjad.Accidental('f'))
            'f'

            >>> str(abjad.Accidental(''))
            ''

            >>> str(abjad.Accidental('qs'))
            'qs'

            >>> str(abjad.Accidental('s'))
            's'

            >>> str(abjad.Accidental('tqs'))
            'tqs'

            >>> str(abjad.Accidental('ss'))
            'ss'

        Returns string.
        """
        if self.semitones in constants._accidental_semitones_to_abbreviation:
            return constants._accidental_semitones_to_abbreviation[self.semitones]
        character = 's'
        if self.semitones < 0:
            character = 'f'
        semitones = abs(self.semitones)
        semitones, remainder = divmod(semitones, 1.0)
        abbreviation = character * int(semitones)
        if remainder:
            abbreviation += 'q{}'.format(character)
        return abbreviation

    def __sub__(self, argument):
        """
        Subtracts `argument` from accidental.

        ..  container:: example

            >>> accidental = abjad.Accidental('qs')

            >>> accidental - accidental
            Accidental('natural')

            >>> accidental - accidental - accidental
            Accidental('quarter flat')

        Returns new accidental.
        """
        if not isinstance(argument, type(self)):
            message = 'can only subtract accidental from other accidental.'
            raise TypeError(message)
        semitones = self.semitones - argument.semitones
        return type(self)(semitones)

    ### PRIVATE METHODS ###

    @classmethod
    def _get_all_accidental_abbreviations(class_):
        return list(constants._accidental_abbreviation_to_symbol.keys())

    @classmethod
    def _get_all_accidental_names(class_):
        return list(constants._accidental_name_to_abbreviation.keys())

    @classmethod
    def _get_all_accidental_semitone_values(class_):
        return list(constants._accidental_semitones_to_abbreviation.keys())

    def _get_format_specification(self):
        return FormatSpecification(
            client=self,
            repr_is_indented=False,
            storage_format_args_values=[self.name],
            storage_format_is_indented=False,
            storage_format_kwargs_names=['arrow'],
            )

    def _get_lilypond_format(self):
        return self._abbreviation

    @classmethod
    def _is_abbreviation(class_, argument):
        if not isinstance(argument, str):
            return False
        return bool(constants._alphabetic_accidental_regex.match(argument))

    @classmethod
    def _is_symbol(class_, argument):
        if not isinstance(argument, str):
            return False
        return bool(constants._symbolic_accidental_regex.match(argument))

    ### PUBLIC PROPERTIES ###

    @property
    def arrow(self):
        """
        Gets arrow of accidental.

        ..  container:: example

            Most accidentals carry no arrow:

            >>> abjad.Accidental('sharp').arrow is None
            True

        ..  container:: example

            Sharp with up-arrow:

            >>> abjad.Accidental('sharp', arrow=abjad.Up).arrow
            Up

            Sharp with down-arrow:

            >>> abjad.Accidental('sharp', arrow=abjad.Down).arrow
            Down

        Arrow property is currently a stub in the object model. You can set the
        property but accidental math and formatting currently ignore the
        setting.

        Returns up, down or none.
        """
        return self._arrow

    @property
    def name(self):
        """
        Gets name of accidental.

        ..  container:: example

            >>> abjad.Accidental('ff').name
            'double flat'

            >>> abjad.Accidental('tqf').name
            'three-quarters flat'

            >>> abjad.Accidental('f').name
            'flat'

            >>> abjad.Accidental('').name
            'natural'

            >>> abjad.Accidental('qs').name
            'quarter sharp'

            >>> abjad.Accidental('s').name
            'sharp'

            >>> abjad.Accidental('tqs').name
            'three-quarters sharp'

            >>> abjad.Accidental('ss').name
            'double sharp'

        Returns string.
        """
        try:
            abbreviation = constants._accidental_semitones_to_abbreviation[self.semitones]
            name = constants._accidental_abbreviation_to_name[abbreviation]
        except KeyError:
            name = str(self)
        return name

    @property
    def semitones(self):
        """
        Gets semitones of accidental.

        ..  container:: example

            >>> abjad.Accidental('ff').semitones
            -2

            >>> abjad.Accidental('tqf').semitones
            -1.5

            >>> abjad.Accidental('f').semitones
            -1

            >>> abjad.Accidental('').semitones
            0

            >>> abjad.Accidental('qs').semitones
            0.5

            >>> abjad.Accidental('s').semitones
            1

            >>> abjad.Accidental('tqs').semitones
            1.5

            >>> abjad.Accidental('ss').semitones
            2

        Returns number.
        """
        return self._semitones

    @property
    def symbol(self):
        """
        Gets symbol of accidental.

        ..  container:: example

            >>> abjad.Accidental('ff').symbol
            'bb'

            >>> abjad.Accidental('tqf').symbol
            'b~'

            >>> abjad.Accidental('f').symbol
            'b'

            >>> abjad.Accidental('').symbol
            ''

            >>> abjad.Accidental('qs').symbol
            '+'

            >>> abjad.Accidental('s').symbol
            '#'

            >>> abjad.Accidental('tqs').symbol
            '#+'

            >>> abjad.Accidental('ss').symbol
            '##'

        Returns string.
        """
        abbreviation = constants._accidental_semitones_to_abbreviation[self.semitones]
        symbol = constants._accidental_abbreviation_to_symbol[abbreviation]
        return symbol

    ### PUBLIC METHODS ###

    # TODO: move somewhere else; remove late imports
    @staticmethod
    def respell_with_flats(selection):
        r"""
        Respells `selection` with flats.

        ..  container:: example

            Respells notes in staff:

            >>> staff = abjad.Staff("c'8 cs'8 d'8 ef'8 e'8 f'8")
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    cs'8
                    d'8
                    ef'8
                    e'8
                    f'8
                }

            >>> abjad.Accidental.respell_with_flats(staff)
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    df'8
                    d'8
                    ef'8
                    e'8
                    f'8
                }

        Returns none.
        """
        import abjad
        for leaf in abjad.iterate(selection).leaves():
            if isinstance(leaf, abjad.Note):
                leaf.written_pitch = leaf.written_pitch._respell_with_flats()
            elif isinstance(leaf, abjad.Chord):
                for note_head in leaf.note_heads:
                    pitch = note_head.written_pitch._respell_with_flats()
                    note_head.written_pitch = pitch

    # TODO: move somewhere else; remove late imports
    @staticmethod
    def respell_with_sharps(selection):
        r"""
        Respells `selection` with sharps.

        ..  container:: example

            Respells notes in staff:

            >>> staff = abjad.Staff("c'8 cs'8 d'8 ef'8 e'8 f'8")
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    cs'8
                    d'8
                    ef'8
                    e'8
                    f'8
                }

            >>> abjad.Accidental.respell_with_sharps(staff)
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    cs'8
                    d'8
                    ds'8
                    e'8
                    f'8
                }

        Returns none.
        """
        import abjad
        for leaf in abjad.iterate(selection).leaves():
            if isinstance(leaf, abjad.Note):
                leaf.written_pitch = leaf.written_pitch._respell_with_sharps()
            elif isinstance(leaf, abjad.Chord):
                for note_head in leaf.note_heads:
                    pitch = note_head.written_pitch._respell_with_sharps()
                    note_head.written_pitch = pitch



import typing
from abjad import enums
from abjad.scheme import Scheme
from abjad.scheme import SchemePair
from abjad.system.AbjadObject import AbjadObject
from abjad.utilities.String import String
from .LilyPondFormatBundle import LilyPondFormatBundle


class LilyPondFormatManager(AbjadObject):
    """
    Manages LilyPond formatting logic.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'LilyPond formatting'

    __slots__ = ()

    indent = 4 * ' '

    ### PRIVATE METHODS ###

    @staticmethod
    def _collect_indicators(component):
        import abjad
        wrappers = []
        parentage = abjad.inspect(component).parentage(include_self=True)
        for parent in parentage:
            wrappers_ = abjad.inspect(parent).wrappers()
            wrappers.extend(wrappers_)
        up_markup_wrappers = []
        down_markup_wrappers = []
        neutral_markup_wrappers = []
        context_wrappers = []
        noncontext_wrappers = []
        # classify wrappers attached to component
        for wrapper in wrappers:
            # skip nonprinting indicators like annotation
            indicator = wrapper.indicator
            if (not hasattr(indicator, '_get_lilypond_format') and
                not hasattr(indicator, '_get_lilypond_format_bundle')
                ):
                continue
            elif wrapper.annotation is not None:
                continue
            # skip comments and commands unless attached directly to us
            elif (wrapper.context is None and
                hasattr(wrapper.indicator, '_format_leaf_children') and
                not getattr(wrapper.indicator, '_format_leaf_children') and
                wrapper.component is not component):
                continue
            # store markup wrappers
            elif isinstance(wrapper.indicator, abjad.Markup):
                if wrapper.indicator.direction is enums.Up:
                    up_markup_wrappers.append(wrapper)
                elif wrapper.indicator.direction is enums.Down:
                    down_markup_wrappers.append(wrapper)
                elif wrapper.indicator.direction in (enums.Center, None):
                    neutral_markup_wrappers.append(wrapper)
            # store context wrappers
            elif wrapper.context is not None:
                if wrapper._is_formattable_for_component(component):
                    context_wrappers.append(wrapper)
            # store noncontext wrappers
            else:
                noncontext_wrappers.append(wrapper)
        indicators = (
            up_markup_wrappers,
            down_markup_wrappers,
            neutral_markup_wrappers,
            context_wrappers,
            noncontext_wrappers,
            )
        return indicators

    @staticmethod
    def _populate_context_setting_format_contributions(component, bundle):
        import abjad
        result = []
        manager = LilyPondFormatManager
        if isinstance(component, abjad.Context):
            for name, value in vars(abjad.setting(component)).items():
                string = manager.format_lilypond_context_setting_in_with_block(
                    name, value)
                result.append(string)
        else:
            contextualizer = abjad.setting(component)
            variables = vars(contextualizer)
            for name, value in variables.items():
                # if we've found a leaf context namespace
                if name.startswith('_'):
                    for x, y in vars(value).items():
                        if not x.startswith('_'):
                            string = \
                                manager.format_lilypond_context_setting_inline(
                                    x, y, name)
                            result.append(string)
                # otherwise we've found a default leaf context setting
                else:
                    # parse default context setting
                    string = manager.format_lilypond_context_setting_inline(
                        name, value)
                    result.append(string)
        result.sort()
        bundle.context_settings.extend(result)

    @staticmethod
    def _populate_context_wrapper_format_contributions(
        component,
        bundle,
        context_wrappers,
        ):
        for wrapper in context_wrappers:
            format_pieces = wrapper._get_format_pieces()
            if isinstance(format_pieces, type(bundle)):
                bundle.update(format_pieces)
            else:
                format_slot = wrapper.indicator._format_slot
                bundle.get(format_slot).indicators.extend(format_pieces)

    @staticmethod
    def _populate_grob_override_format_contributions(component, bundle):
        import abjad
        result = []
        once = isinstance(component, abjad.Leaf)
        grob = abjad.override(component)
        contributions = grob._list_format_contributions(
            'override',
            once=once,
            )
        for string in result[:]:
            if 'NoteHead' in string and 'pitch' in string:
                contributions.remove(string)
        try:
            written_pitch = component.written_pitch
            arrow = written_pitch.arrow
        except AttributeError:
            arrow = None
        if arrow in (enums.Up, enums.Down):
            contributions_ = written_pitch._list_format_contributions()
            contributions.extend(contributions_)
        bundle.grob_overrides.extend(contributions)

    @staticmethod
    def _populate_grob_revert_format_contributions(component, bundle):
        import abjad
        if not isinstance(component, abjad.Leaf):
            manager = abjad.override(component)
            contributions = manager._list_format_contributions('revert')
            bundle.grob_reverts.extend(contributions)

    @staticmethod
    def _populate_indicator_format_contributions(component, bundle):
        manager = LilyPondFormatManager
        (
            up_markup_wrappers,
            down_markup_wrappers,
            neutral_markup_wrappers,
            context_wrappers,
            noncontext_wrappers,
            ) = LilyPondFormatManager._collect_indicators(component)
        manager._populate_markup_format_contributions(
            component,
            bundle,
            up_markup_wrappers,
            down_markup_wrappers,
            neutral_markup_wrappers,
            )
        manager._populate_context_wrapper_format_contributions(
            component,
            bundle,
            context_wrappers,
            )
        manager._populate_noncontext_wrapper_format_contributions(
            component,
            bundle,
            noncontext_wrappers,
            )

    @staticmethod
    def _populate_markup_format_contributions(
        component,
        bundle,
        up_markup_wrappers,
        down_markup_wrappers,
        neutral_markup_wrappers,
        ):
        import abjad
        for wrappers in (
            up_markup_wrappers,
            down_markup_wrappers,
            neutral_markup_wrappers,
            ):
            for wrapper in wrappers:
                if wrapper.indicator.direction is None:
                    markup = abjad.Markup(wrapper.indicator, direction='-')
                else:
                    markup = wrapper.indicator
                format_pieces = markup._get_format_pieces()
                format_pieces = LilyPondFormatManager.tag(
                    format_pieces,
                    wrapper.tag,
                    deactivate=wrapper.deactivate,
                    )
                bundle.after.markup.extend(format_pieces)

    @staticmethod
    def _populate_noncontext_wrapper_format_contributions(
        component,
        bundle,
        noncontext_wrappers,
        ):
        for wrapper in noncontext_wrappers:
            indicator = wrapper.indicator
            if hasattr(indicator, '_get_lilypond_format_bundle'):
                bundle_ = indicator._get_lilypond_format_bundle()
                if wrapper.tag:
                    bundle_.tag_format_contributions(
                        wrapper.tag,
                        deactivate=wrapper.deactivate,
                        )
                if bundle_ is not None:
                    bundle.update(bundle_)

    @staticmethod
    def _populate_spanner_format_contributions(component, bundle):
        import abjad
        if not hasattr(component, '_spanners'):
            return
        pairs = []
        for spanner in abjad.inspect(component).spanners():
            spanner_bundle = spanner._get_lilypond_format_bundle(component)
            spanner_bundle.tag_format_contributions(
                spanner._tag,
                deactivate=spanner._deactivate,
                )
            pair = (spanner, spanner_bundle)
            pairs.append(pair)
        pairs.sort(key=lambda _: type(_[0]).__name__)
        for spanner, spanner_bundle in pairs:
            bundle.update(spanner_bundle)

    ### PUBLIC METHODS ###

    @staticmethod
    def align_tags(string, n) -> str:
        """
        Line-breaks ``string`` and aligns tags starting a column ``n``.
        """
        assert isinstance(n, int), repr(n)
        lines = []
        for line in string.split('\n'):
            if '%!' not in line:
                lines.append(line)
                continue
            location = line.find('%!')
            left = line[:location].rstrip()
            right = line[location:]
            pad = n - len(left)
            if pad < 1:
                pad = 1
            line = left + pad * ' ' + right
            lines.append(line)
        string = '\n'.join(lines)
        return string

    @staticmethod
    def bundle_format_contributions(component) -> LilyPondFormatBundle:
        """
        Gets all format contributions for ``component``.
        """
        manager = LilyPondFormatManager
        bundle = LilyPondFormatBundle()
        manager._populate_indicator_format_contributions(component, bundle)
        manager._populate_spanner_format_contributions(component, bundle)
        manager._populate_context_setting_format_contributions(
            component, bundle)
        manager._populate_grob_override_format_contributions(component, bundle)
        manager._populate_grob_revert_format_contributions(component, bundle)
        bundle.sort_overrides()
        return bundle

    @staticmethod
    def format_lilypond_attribute(attribute) -> str:
        """
        Formats LilyPond attribute according to Scheme formatting conventions.
        """
        assert isinstance(attribute, str), repr(attribute)
        attribute = attribute.replace('__', ".")
        result = attribute.replace('_', '-')
        return result

    @staticmethod
    def format_lilypond_context_setting_in_with_block(name, value) -> str:
        """
        Formats LilyPond context setting ``name`` with ``value`` in LilyPond
        with-block.
        """
        assert isinstance(name, str), repr(name)
        name = name.split('_')
        first = name[0:1]
        rest = name[1:]
        rest = [x.title() for x in rest]
        name = first + rest
        name = ''.join(name)
        value = LilyPondFormatManager.format_lilypond_value(value)
        value_parts = value.split('\n')
        result = rf'{name!s} = {value_parts[0]!s}'
        pieces = [result]
        for part in value_parts[1:]:
            pieces.append(LilyPondFormatManager.indent + part)
        return '\n'.join(pieces)

    @staticmethod
    def format_lilypond_context_setting_inline(
        name,
        value,
        context=None,
        ) -> str:
        """
        Formats LilyPond context setting ``name`` with ``value`` in
        ``context``.
        """
        name = name.split('_')
        first = name[0:1]
        rest = name[1:]
        rest = [x.title() for x in rest]
        name = first + rest
        name = ''.join(name)
        value = LilyPondFormatManager.format_lilypond_value(value)
        if context is not None:
            context_string = context[1:]
            context_string = context_string.split('_')
            context_string = [x.title() for x in context_string]
            context_string = ''.join(context_string)
            context_string += '.'
        else:
            context_string = ''
        result = rf'\set {context_string}{name} = {value}'
        return result

    @staticmethod
    def format_lilypond_value(argument) -> str:
        """
        Formats LilyPond ``argument`` according to Scheme formatting
        conventions.
        """
        if ('_get_lilypond_format' in dir(argument) and 
            not isinstance(argument, str)):
            pass
        elif argument in (True, False):
            argument = Scheme(argument)
        elif argument in (enums.Up, enums.Down, enums.Left, enums.Right, enums.Center):
            argument = Scheme(repr(argument).lower())
        elif isinstance(argument, int) or isinstance(argument, float):
            argument = Scheme(argument)
        elif argument in Scheme.lilypond_color_constants:
            argument = Scheme(argument)
        elif isinstance(argument, str) and argument.startswith('#'):
            argument = Scheme(argument)
        elif isinstance(argument, str) and '::' in argument:
            argument = Scheme(argument)
        elif isinstance(argument, tuple) and len(argument) == 2:
            argument = SchemePair(argument)
        elif isinstance(argument, str) and ' ' not in argument:
            argument = Scheme(argument, quoting="'")
        elif isinstance(argument, str) and ' ' in argument:
            argument = Scheme(argument)
        else:
            argument = Scheme(argument, quoting="'")
        return format(argument, 'lilypond')

    @staticmethod
    def left_shift_tags(text, realign=None) -> str:
        """
        Left shifts tags in ``strings`` and realigns to column ``realign``.
        """
        strings = text.split('\n')
        strings_ = [] 
        for string in strings:
            if '%@% ' not in string or '%!' not in string:
                strings_.append(string)
                continue
            if not string.startswith(4 * ' '):
                strings_.append(string)
                continue
            string_ = string[4:]
            tag_start = string_.find('%!')
            string_ = list(string_)
            string_[tag_start:tag_start] = 4 * ' '
            string_ = ''.join(string_)
            strings_.append(string_)
        text = '\n'.join(strings_)
        if realign is not None:
            text = LilyPondFormatManager.align_tags(text, n=realign)
        return text

    @staticmethod
    def make_lilypond_override_string(
        grob,
        attribute,
        value,
        context=None,
        once=False,
        ) -> str:
        """
        Makes Lilypond override string.
        """
        grob = String(grob).to_upper_camel_case()
        attribute = LilyPondFormatManager.format_lilypond_attribute(attribute)
        value = LilyPondFormatManager.format_lilypond_value(value)
        if context is not None:
            context = String(context).capitalize_start() + '.'
        else:
            context = ''
        if once is True:
            once = r'\once '
        else:
            once = ''
        result = rf'{once}\override {context}{grob}.{attribute} = {value}'
        return result

    @staticmethod
    def make_lilypond_revert_string(grob, attribute, context=None) -> str:
        r"""
        Makes LilyPond revert string.

        ..  container:: example

            >>> abjad.LilyPondFormatManager.make_lilypond_revert_string(
            ...     'glissando',
            ...     'bound_details__right__arrow',
            ...     )
            '\\revert Glissando.bound-details.right.arrow'

        """
        grob = String(grob).to_upper_camel_case()
        dotted = LilyPondFormatManager.format_lilypond_attribute(attribute)
        if context is not None:
            context = String(context).to_upper_camel_case()
            context += '.'
        else:
            context = ''
        result = rf'\revert {context}{grob}.{dotted}'
        return result

    @staticmethod
    def make_lilypond_tweak_string(
        attribute,
        value,
        directed=True,
        grob=None,
        ) -> str:
        r"""
        Makes Lilypond \tweak string.

        Returns string.
        """
        if grob is not None:
            grob = String(grob).to_upper_camel_case()
            grob += '.'
        else:
            grob = ''
        attribute = LilyPondFormatManager.format_lilypond_attribute(attribute)
        value = LilyPondFormatManager.format_lilypond_value(value)
        string = rf'\tweak {grob}{attribute} {value}'
        if directed:
            string = '- ' + string
        return string

    @staticmethod
    def report_spanner_format_contributions(spanner) -> str:
        """
        Reports spanner format contributions for every leaf in ``spanner``.

        ..  container:: example

            >>> staff = abjad.Staff("c8 d e f")
            >>> spanner = abjad.Beam()
            >>> abjad.attach(spanner, staff[:])

            >>> manager = abjad.LilyPondFormatManager
            >>> print(manager.report_spanner_format_contributions(spanner))
            c8	abjad.LilyPondFormatBundle(
                    after=abjad.SlotContributions(
                        spanner_starts=['['],
                        ),
                    )
            d8	abjad.LilyPondFormatBundle()
            e8	abjad.LilyPondFormatBundle()
            f8	abjad.LilyPondFormatBundle(
                    after=abjad.SlotContributions(
                        spanner_stops=[']'],
                        ),
                    )

        Returns string or none.
        """
        strings = []
        for leaf in spanner.leaves:
            bundle = spanner._get_lilypond_format_bundle(leaf)
            bundle_pieces = format(bundle).split('\n')
            strings.append(f'{leaf!s}\t{bundle_pieces[0]}')
            for piece in bundle_pieces[1:]:
                strings.append(f'\t{piece}')
        string = '\n'.join(strings)
        return string

    @staticmethod
    def tag(strings, tag, deactivate=None) -> typing.List[str]:
        """
        Tags ``strings`` with ``tag``.
        """
        if not tag:
            return strings
        if not strings:
            return strings
        if deactivate is not None:
            assert isinstance(deactivate, type(True)), repr(deactivate)
        length = max([len(_) for _ in strings])
        strings_ = []
        for string in strings:
            if '%!' not in string:
                pad = length - len(string)
            else:
                pad = 0
            tag_ = pad * ' ' + ' ' + '%!' + ' ' + str(tag)
            string = string + tag_
            strings_.append(string)
        if deactivate is True:
            strings_ = ['%@% ' + _ for _ in strings_]
        return strings_



#! /usr/bin/env python
import abjad
import ide
import os
import pathlib
import sys
import time
import traceback


if __name__ == '__main__':

    try:
        from definition import maker
        from __metadata__ import metadata as metadata
        {previous_segment_metadata_import_statement}
    except ImportError:
        traceback.print_exc()
        sys.exit(1)

    try:
        segment_directory = ide.Path(os.path.realpath(__file__)).parent
        scores_directory = segment_directory.parent.parent.parent.parent
        segment_directory = ide.Path(
            segment_directory,
            scores=scores_directory,
            )
        assert segment_directory.is_score_package_path(), repr(segment_directory)
        illustration_ly = segment_directory('illustration.ly')
        print(' Running segment-maker ...')
        with abjad.Timer() as timer:
            lilypond_file = maker.run(
                metadata=metadata,
                previous_metadata=previous_metadata,
                segment_directory=segment_directory,
                )
        segment_maker_runtime = int(timer.elapsed_time)
        count = segment_maker_runtime
        counter = abjad.String('second').pluralize(count)
        message = f' Segment-maker runtime {{count}} {{counter}} ...'
        print(message)
        segment_maker_runtime = (count, counter)
        segment_directory.write_metadata_py(maker.metadata)
        result = abjad.persist(lilypond_file).as_ly(illustration_ly, strict=89)
        abjad_format_time = int(result[1])
        count = abjad_format_time
        counter = abjad.String('second').pluralize(count)
        message = f' Abjad format time {{count}} {{counter}} ...'
        print(message)
        abjad_format_time = (count, counter)
    except:
        traceback.print_exc()
        sys.exit(1)

    try:
        if 'GlobalSkips' in lilypond_file:
            context = lilypond_file['GlobalSkips']
            measure_count = len(context)
            counter = abjad.String('measure').pluralize(measure_count)
            message = f' Wrote {{measure_count}} {{counter}}'
            message += f' to {{illustration_ly.trim()}} ...'
            print(message)
            time_signatures = []
            prototype = abjad.TimeSignature
            for skip in context:
                time_signature = abjad.inspect(skip).effective(prototype)
                assert isinstance(time_signature, prototype)
                time_signatures.append(str(time_signature))
        else:
            measure_count = None
            time_signatures = None
    except:
        traceback.print_exc()
        sys.exit(1)

    try:
        text = illustration_ly.read_text()
        text = abjad.LilyPondFormatManager.left_shift_tags(text, realign=99)
        illustration_ly.write_text(text)
        for job in [
            abjad.Job.handle_edition_tags(illustration_ly),
            abjad.Job.show_music_annotations(illustration_ly, undo=True),
            abjad.Job.handle_fermata_bar_lines(segment_directory),
            abjad.Job.handle_shifted_clefs(segment_directory),
            ]:
            for message in job():
                print(' ' + message)
    except:
        traceback.print_exc()
        sys.exit(1)

    try:
        layout_py = segment_directory('layout.py')
        if not layout_py.exists():
            print(' Writing stub layout.py ...')
            layout_py.write_text('')
        layout_ly = segment_directory('layout.ly')
        if not layout_ly.exists():
            print(' Writing stub layout.ly ...')
            layout_ly.write_text('')
    except:
        traceback.print_exc()
        sys.exit(1)

    try:
        if layout_py.read_text() == '':
            empty_layout = True
        else:
            empty_layout = False
        if empty_layout:
            print(f' Ignoring empty {{layout_py.trim()}} ...')
        else:
            layout_time_signatures = layout_ly.get_preamble_time_signatures()
            if layout_time_signatures is not None:
                assert isinstance(layout_time_signatures, list)
                layout_measure_count = len(layout_time_signatures)
                counter = abjad.String('measure').pluralize(
                    layout_measure_count)
                message = f' Found {{layout_measure_count}} {{counter}}'
                message += f' in {{layout_ly.trim()}} ...'
                print(message)
                if layout_time_signatures == time_signatures:
                    message = ' Music time signatures match'
                    message += ' layout time signatures ...'
                    print(message)
                else:
                    message = 'Music time signatures do not match'
                    message += ' layout time signatures ...'
                    print(message)
                    print(f' Remaking {{layout_ly.trim()}} ...')
                    ide = ide.AbjadIDE()
                    ide._make_layout_ly(layout_py)
                    counter = abjad.String('measure').pluralize(measure_count)
                    message = f' Found {{measure_count}} {{counter}}'
                    message += f' in {{illustration_ly.trim()}} ...'
                    print(message)
                    layout_time_signatures = \
                        layout_ly.get_preamble_time_signatures()
                    layout_measure_count = len(layout_time_signatures)
                    counter = abjad.String('measure').pluralize(
                        layout_measure_count
                        )
                    message = f' Found {{layout_measure_count}} {{counter}}'
                    message += f' in {{layout_ly.trim()}} ...'
                    print(message)
                    if layout_time_signatures != time_signatures:
                        message = ' Music time signatures still do not match'
                        message += ' layout time signatures ...'
                        print(message)
    except:
        traceback.print_exc()
        sys.exit(1)

    try:
        if getattr(maker, 'do_not_externalize', False) is not True:
            illustration_ly.extern()
            illustration_ily = illustration_ly.with_suffix('.ily')
            assert illustration_ily.is_file()
        with abjad.Timer() as timer:
            print(' Running LilyPond ...')
            abjad.IOManager.run_lilypond(illustration_ly)
        lilypond_runtime = int(timer.elapsed_time)
        count = lilypond_runtime
        counter = abjad.String('second').pluralize(count)
        message = f' LilyPond runtime {{count}} {{counter}} ...'
        print(message)
        lilypond_runtime = (count, counter)
    except:
        traceback.print_exc()
        sys.exit(1)

    try:
        history = segment_directory('.history')
        with history.open(mode='a') as pointer:
            pointer.write('\n')
            line = time.strftime('%Y-%m-%d %H:%M:%S') + '\n'
            pointer.write(line)
            count, counter = segment_maker_runtime
            line = f' Segment-maker runtime: {{count}} {{counter}}\n'
            pointer.write(line)
            count, counter = abjad_format_time
            line = f' Abjad format time: {{count}} {{counter}}\n'
            pointer.write(line)
            count, counter = lilypond_runtime
            line = f' LilyPond runtime: {{count}} {{counter}}\n'
            pointer.write(line)
    except:
        traceback.print_exc()
        sys.exit(1)

    sys.exit(0)
